# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
from unittest.mock import patch

import pytest

from otx.api.configuration import ConfigurableParameters
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.inference_parameters import InferenceParameters
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model import ModelConfiguration, ModelEntity, ModelPrecision
from otx.api.entities.optimization_parameters import OptimizationParameters
from otx.api.entities.resultset import ResultSetEntity
from otx.api.entities.train_parameters import TrainParameters
from otx.api.usecases.tasks.interfaces.evaluate_interface import IEvaluationTask
from otx.api.usecases.tasks.interfaces.export_interface import ExportType, IExportTask
from otx.api.usecases.tasks.interfaces.inference_interface import (
    IInferenceTask,
    IRawInference,
)
from otx.api.usecases.tasks.interfaces.optimization_interface import (
    IOptimizationTask,
    OptimizationType,
)
from otx.api.usecases.tasks.interfaces.training_interface import ITrainingTask
from otx.api.usecases.tasks.interfaces.unload_interface import IUnload
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestIEvaluationTask:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    @patch(
        "otx.api.usecases.tasks.interfaces.evaluate_interface.IEvaluationTask.__abstractmethods__",
        set(),
    )
    def test_evaluate_interface(self):
        """
        <b>Description:</b>
        Check IEvaluationTask class object initialization

        <b>Input data:</b>
        IEvaluationTask object

        <b>Expected results:</b>
        Test passes if IEvaluationTask object evaluate method raises NotImplementedError exception
        """
        dataset = DatasetEntity()
        configuration = ModelConfiguration(
            configurable_parameters=ConfigurableParameters(header="Test Header"),
            label_schema=LabelSchemaEntity(),
        )
        model_entity = ModelEntity(configuration=configuration, train_dataset=dataset)
        with pytest.raises(NotImplementedError):
            IEvaluationTask().evaluate(
                ResultSetEntity(
                    model=model_entity,
                    ground_truth_dataset=dataset,
                    prediction_dataset=dataset,
                )
            )


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestIInferenceTask:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    @patch(
        "otx.api.usecases.tasks.interfaces.inference_interface.IInferenceTask.__abstractmethods__",
        set(),
    )
    def test_i_inference_task(self):
        """
        <b>Description:</b>
        Check IInferenceTask class object initialization

        <b>Input data:</b>
        IInferenceTask object

        <b>Expected results:</b>
        Test passes if IInferenceTask object infer method raises NotImplementedError exception
        """
        dataset = DatasetEntity()
        inference_parameters = InferenceParameters()
        with pytest.raises(NotImplementedError):
            IInferenceTask().infer(dataset=dataset, inference_parameters=inference_parameters)


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestIRawInference:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    @patch(
        "otx.api.usecases.tasks.interfaces.inference_interface.IRawInference.__abstractmethods__",
        set(),
    )
    def test_i_raw_inference(self):
        """
        <b>Description:</b>
        Check TestIRawInference class object initialization

        <b>Input data:</b>
        TestIRawInference object

        <b>Expected results:</b>
        Test passes if TestIRawInference object raw_infer method raises NotImplementedError exception
        """
        with pytest.raises(NotImplementedError):
            IRawInference().raw_infer(input_tensors={}, output_tensors={})


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestOptimizationType:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_optimization_type(self):
        """
        <b>Description:</b>
        Check OptimizationType Enum class elements

        <b>Expected results:</b>
        Test passes if OptimizationType Enum class length is equal to expected value and its elements have expected
        sequence numbers
        """
        assert len(OptimizationType) == 2
        assert OptimizationType.POT.value == 1
        assert OptimizationType.NNCF.value == 2


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestIOptimizationTask:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    @patch(
        "otx.api.usecases.tasks.interfaces.optimization_interface.IOptimizationTask.__abstractmethods__",
        set(),
    )
    def test_optimization_interface(self):
        """
        <b>Description:</b>
        Check IOptimizationTask class object initialization

        <b>Input data:</b>
        IOptimizationTask object

        <b>Expected results:</b>
        Test passes if IOptimizationTask object optimize method raises NotImplementedError exception
        """
        dataset = DatasetEntity()
        configuration = ModelConfiguration(
            configurable_parameters=ConfigurableParameters(header="Test Header"),
            label_schema=LabelSchemaEntity(),
        )
        model_entity = ModelEntity(configuration=configuration, train_dataset=dataset)
        optimization_parameters = OptimizationParameters()
        with pytest.raises(NotImplementedError):
            IOptimizationTask().optimize(
                optimization_type=OptimizationType.POT,
                dataset=dataset,
                output_model=model_entity,
                optimization_parameters=optimization_parameters,
            )


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestITrainingTask:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    @patch(
        "otx.api.usecases.tasks.interfaces.training_interface.ITrainingTask.__abstractmethods__",
        set(),
    )
    def test_training_interface(self):
        """
        <b>Description:</b>
        Check ITrainingTask class object initialization

        <b>Input data:</b>
        ITrainingTask object

        <b>Expected results:</b>
        Test passes if ITrainingTask object methods raise NotImplementedError exception
        """
        i_training_task = ITrainingTask()
        dataset = DatasetEntity()
        configuration = ModelConfiguration(
            configurable_parameters=ConfigurableParameters(header="Test Header"),
            label_schema=LabelSchemaEntity(),
        )
        model_entity = ModelEntity(configuration=configuration, train_dataset=dataset)
        train_parameters = TrainParameters()

        with pytest.raises(NotImplementedError):
            i_training_task.save_model(model_entity)
        with pytest.raises(NotImplementedError):
            i_training_task.train(
                dataset=dataset,
                output_model=model_entity,
                train_parameters=train_parameters,
            )
        with pytest.raises(NotImplementedError):
            i_training_task.cancel_training()


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestIUnload:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    @patch(
        "otx.api.usecases.tasks.interfaces.unload_interface.IUnload.__abstractmethods__",
        set(),
    )
    def test_unload_interface(self):
        """
        <b>Description:</b>
        Check IUnload class object initialization

        <b>Input data:</b>
        IUnload object

        <b>Expected results:</b>
        Test passes if IUnload object unload method raises NotImplementedError exception
        """
        with pytest.raises(NotImplementedError):
            IUnload().unload()


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestExportType:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_export_type(self):
        """
        <b>Description:</b>
        Check ExportType Enum class elements

        <b>Expected results:</b>
        Test passes if ExportType Enum class length is equal to expected value and its elements have expected
        sequence numbers
        """
        assert len(ExportType) == 1
        assert ExportType.OPENVINO.value == 1


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestIExportTask:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    @patch(
        "otx.api.usecases.tasks.interfaces.export_interface.IExportTask.__abstractmethods__",
        set(),
    )
    def test_export_interface(self):
        """
        <b>Description:</b>
        Check IExportTask class object initialization

        <b>Input data:</b>
        IExportTask object

        <b>Expected results:</b>
        Test passes if IExportTask object export method raises NotImplementedError exception
        """
        dataset = DatasetEntity()
        configuration = ModelConfiguration(
            configurable_parameters=ConfigurableParameters(header="Test Header"),
            label_schema=LabelSchemaEntity(),
        )
        model_entity = ModelEntity(configuration=configuration, train_dataset=dataset)
        with pytest.raises(NotImplementedError):
            IExportTask().export(
                export_type=ExportType.OPENVINO,
                output_model=model_entity,
                precision=ModelPrecision.FP32,
                dump_features=False,
            )



"""Collection of utils for task implementation in Detection Task."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from .data import (
    format_list_to_str,
    get_anchor_boxes,
    get_sizes_from_dataset_entity,
    load_dataset_items_coco_format,
)
from .utils import generate_label_schema, get_det_model_api_configuration

__all__ = [
    "get_det_model_api_configuration",
    "load_dataset_items_coco_format",
    "get_sizes_from_dataset_entity",
    "get_anchor_boxes",
    "format_list_to_str",
    "generate_label_schema",
]



"""Module for noisy label detection features."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .base import LossDynamicsTracker, LossDynamicsTrackingMixin
from .loss_dynamics_tracking_hook import LossDynamicsTrackingHook

__all__ = ["LossDynamicsTrackingHook", "LossDynamicsTracker", "LossDynamicsTrackingMixin"]



"""Neck list of mmdetection adapters."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .mmov_fpn import MMOVFPN
from .mmov_ssd_neck import MMOVSSDNeck
from .mmov_yolov3_neck import MMOVYOLOV3Neck

__all__ = [
    "MMOVFPN",
    "MMOVSSDNeck",
    "MMOVYOLOV3Neck",
]



"""Hook module to track loss dynamics during training and export these statistics to Datumaro format."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from collections import defaultdict
from typing import Any, Dict, List

from otx.api.entities.datasets import DatasetEntity

__all__ = ["LossDynamicsTracker", "LossDynamicsTrackingMixin"]


class LossDynamicsTracker:
    """Class to track loss dynamics and export it to Datumaro format."""

    def __init__(self) -> None:
        self.initialized = False

    def init_with_otx_dataset(self, otx_dataset: DatasetEntity) -> None:
        """DatasetEntity should be injected to the tracker for the initialization."""
        self._loss_dynamics: Dict[Any, List] = defaultdict(list)
        self.initialized = True

    def accumulate(self, outputs, iter) -> None:
        """Accumulate training loss dynamics for each training step."""
        raise NotImplementedError()

    def export(self, output_path: str) -> None:
        """Export loss dynamics statistics to Datumaro format."""
        raise NotImplementedError()


class LossDynamicsTrackingMixin:
    """Mix-in to track loss dynamics during training."""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        self._loss_dyns_tracker = LossDynamicsTracker()

    @property
    def loss_dyns_tracker(self) -> LossDynamicsTracker:
        """Get tracker."""
        return self._loss_dyns_tracker



"""Tiling Module."""

# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import copy
from itertools import product
from typing import Any, Dict, List, Tuple, Union

import numpy as np
from openvino.model_zoo.model_api.models import Model

from otx.api.utils.async_pipeline import OTXDetectionAsyncPipeline
from otx.api.utils.detection_utils import detection2array
from otx.api.utils.nms import multiclass_nms


class Tiler:
    """Tile Image into (non)overlapping Patches. Images are tiled in order to efficiently process large images.

    Args:
        tile_size: Tile dimension for each patch
        overlap: Overlap between adjacent tile
        max_number: max number of prediction per image
        detector: OpenVINO adaptor model
        classifier: Tile classifier OpenVINO adaptor model
        segm: enable instance segmentation mask output
        mode: async or sync mode
    """

    def __init__(
        self,
        tile_size: int,
        overlap: float,
        max_number: int,
        detector: Any,
        classifier: Model,
        segm: bool = False,
        mode: str = "async",
    ):  # pylint: disable=too-many-arguments
        self.tile_size = tile_size
        self.overlap = overlap
        self.max_number = max_number
        self.model = detector
        self.classifier = classifier
        self.segm = segm
        if self.segm:
            self.model.disable_mask_resizing()
        if mode == "async":
            self.async_pipeline = OTXDetectionAsyncPipeline(self.model)

    def tile(self, image: np.ndarray) -> List[List[int]]:
        """Tiles an input image to either overlapping, non-overlapping or random patches.

        Args:
            image: Input image to tile.

        Returns:
            Tiles coordinates
        """
        height, width = image.shape[:2]

        coords = [[0, 0, width, height]]
        for (loc_j, loc_i) in product(
            range(0, width - self.tile_size + 1, int(self.tile_size * (1 - self.overlap))),
            range(0, height - self.tile_size + 1, int(self.tile_size * (1 - self.overlap))),
        ):
            coords.append([loc_j, loc_i, loc_j + self.tile_size, loc_i + self.tile_size])
        return coords

    def filter_tiles_by_objectness(
        self, image: np.ndarray, tile_coords: List[List[int]], confidence_threshold: float = 0.35
    ):
        """Filter tiles by objectness score by running tile classifier.

        Args:
            image (np.ndarray): full size image
            tile_coords (List[List[int]]): tile coordinates

        Returns:
            keep_coords: tile coordinates to keep
        """
        keep_coords = []
        for i, coord in enumerate(tile_coords):
            tile_img = self.crop_tile(image, coord)
            tile_dict, _ = self.model.preprocess(tile_img)
            objectness_score = self.classifier.infer_sync(tile_dict)
            if i == 0 or objectness_score["tile_prob"] > confidence_threshold:
                keep_coords.append(coord)
        return keep_coords

    def predict(self, image: np.ndarray, mode: str = "async"):
        """Predict by cropping full image to tiles.

        Args:
            image (np.ndarray): full size image

        Returns:
            detection: prediction results
            features: saliency map and feature vector
        """
        tile_coords = self.tile(image)
        if isinstance(self.classifier, Model):
            tile_coords = self.filter_tiles_by_objectness(image, tile_coords)

        if mode == "sync":
            return self.predict_sync(image, tile_coords)
        return self.predict_async(image, tile_coords)

    def predict_sync(self, image: np.ndarray, tile_coords: List[List[int]]):
        """Predict by cropping full image to tiles synchronously.

        Args:
            image (np.ndarray): full size image
            tile_coords (List[List[int]]): tile coordinates

        Returns:
            detection: prediction results
            features: saliency map and feature vector
        """
        features = (None, None)
        tile_results = []

        for i, coord in enumerate(tile_coords):
            tile_img = self.crop_tile(image, coord)
            tile_dict, tile_meta = self.model.preprocess(tile_img)
            raw_predictions = self.model.infer_sync(tile_dict)
            predictions = self.model.postprocess(raw_predictions, tile_meta)
            tile_result = self.postprocess_tile(predictions, *coord[:2])
            # cache full image feature vector and saliency map at 0 index
            if i == 0 and ("feature_vector" in raw_predictions or "saliency_map" in raw_predictions):
                features = (
                    copy.deepcopy(raw_predictions["feature_vector"].reshape(-1)),
                    copy.deepcopy(raw_predictions["saliency_map"][0]),
                )

            tile_results.append(tile_result)

        results = self.merge_results(tile_results, image.shape)
        return results, features

    def predict_async(self, image: np.ndarray, tile_coords: List[List[int]]):
        """Predict by cropping full image to tiles asynchronously.

        Args:
            image (np.ndarray): full size image
            tile_coords (List[List[int]]): tile coordinates

        Returns:
            detection: prediction results
            features: saliency map and feature vector
        """
        num_tiles = len(tile_coords)

        processed_tiles = 0
        tile_results = []
        features = (None, None)
        for i, coord in enumerate(tile_coords):
            pred = self.async_pipeline.get_result(processed_tiles)
            while pred:
                tile_prediction, meta, feats = pred
                if meta["tile_i"] == 0:
                    features = feats
                tile_result = self.postprocess_tile(tile_prediction, *meta["coord"][:2])
                tile_results.append(tile_result)
                processed_tiles += 1
                pred = self.async_pipeline.get_result(processed_tiles)
            self.async_pipeline.submit_data(self.crop_tile(image, coord), i, {"coord": coord, "tile_i": i})

        self.async_pipeline.await_all()
        for j in range(processed_tiles, num_tiles):
            tile_prediction, meta, feats = self.async_pipeline.get_result(j)
            if meta["tile_i"] == 0:
                features = feats
            tile_result = self.postprocess_tile(tile_prediction, *meta["coord"][:2])
            tile_results.append(tile_result)
        assert j == num_tiles - 1, "Number of tiles processed does not match number of tiles"
        return self.merge_results(tile_results, image.shape), features

    def postprocess_tile(self, predictions: Union[List, Tuple], offset_x: int, offset_y: int) -> Dict[str, List]:
        """Postprocess single tile prediction.

        Args:
            predictions (Union[List, Tuple]): predictions from model
            offset_x (int): tile offset in x direction
            offset_y (int): tile offset in y direction

        Returns:
            Dict[str, List]: postprocessed predictions - bboxes and masks
        """
        output_dict: dict = {"bboxes": [], "masks": []}
        if self.segm:
            tile_scores, tile_labels, tile_boxes, tile_masks = predictions
            tile_boxes += np.tile([offset_x, offset_y], 2)
            out = np.concatenate(
                (
                    tile_labels[:, np.newaxis],
                    tile_scores[:, np.newaxis],
                    tile_boxes,
                ),
                -1,
            )
            output_dict["masks"] = tile_masks
        else:
            assert isinstance(predictions, list)
            out = detection2array(predictions)
            out[:, 2:] += np.tile([offset_x, offset_y], 2)
        output_dict["bboxes"] = out
        return output_dict

    def crop_tile(self, image: np.ndarray, coord: List[int]) -> np.ndarray:
        """Crop tile from full image.

        Args:
            image (np.ndarray): full-res image
            coord (List): tile coordinates

        Returns:
            np.ndarray: cropped tile
        """
        x1, y1, x2, y2 = coord
        return image[y1:y2, x1:x2]

    @staticmethod
    def detection2tuple(detections: np.ndarray):
        """Convert detection to tuple.

        Args:
            detections (np.ndarray): prediction results in numpy array

        Returns:
            scores (np.ndarray): scores between 0-1
            labels (np.ndarray): label indices
            boxes (np.ndarray): boxes
        """
        labels = detections[:, 0]
        scores = detections[:, 1]
        boxes = detections[:, 2:]
        return scores, labels, boxes

    def merge_results(self, results: List[Dict], shape: List[int]):
        """Merge results from tiles.

        Args:
            results (List[Dict]): list of tile results
            shape (List[int]): original full-res image shape
        """

        detections = np.empty((0, 6), dtype=np.float32)
        masks = []
        for result in results:
            if len(result["bboxes"]):
                detections = np.concatenate((detections, result["bboxes"]))
                if self.segm:
                    masks.extend(result["masks"])

        if np.prod(detections.shape):
            detections, keep = multiclass_nms(detections, max_num=self.max_number)
            if self.segm:
                masks = [masks[keep_idx] for keep_idx in keep]
                self.resize_masks(masks, detections, shape)
                detections = *Tiler.detection2tuple(detections), masks
        return detections

    def resize_masks(self, masks: List, dets: np.ndarray, shape: List[int]):
        """Resize Masks.

        Args:
            masks (List): list of raw np.ndarray masks
            dets (np.ndarray): detections including labels, scores, and boxes
            shape (List[int]): original full-res image shape
        """
        for i, (det, mask) in enumerate(zip(dets, masks)):
            masks[i] = self.model.segm_postprocess(det[2:], mask, *shape[:-1])



"""OTX Algorithms - Classification Losses."""

# Copyright (C) 2023 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from .asymmetric_angular_loss_with_ignore import AsymmetricAngularLossWithIgnore
from .asymmetric_loss_with_ignore import AsymmetricLossWithIgnore
from .barlowtwins_loss import BarlowTwinsLoss
from .cross_entropy_loss import CrossEntropyLossWithIgnore
from .ib_loss import IBLoss

__all__ = [
    "AsymmetricAngularLossWithIgnore",
    "AsymmetricLossWithIgnore",
    "BarlowTwinsLoss",
    "CrossEntropyLossWithIgnore",
    "IBLoss",
]



"""Unit-Test case for otx.core.data.adapter.detection_dataset_adapter."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#
import os

from otx.api.entities.annotation import NullAnnotationSceneEntity
from otx.api.entities.datasets import DatasetEntity
from otx.api.entities.label_schema import LabelSchemaEntity
from otx.api.entities.model_template import TaskType
from otx.api.entities.subset import Subset
from otx.core.data.adapter.detection_dataset_adapter import DetectionDatasetAdapter
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.core.data.test_helpers import (
    TASK_NAME_TO_DATA_ROOT,
    TASK_NAME_TO_TASK_TYPE,
)


class TestOTXDetectionDatasetAdapter:
    def setup_method(self):
        self.root_path = os.getcwd()

    @e2e_pytest_unit
    def test_detection(self):
        task = "detection"

        task_type: TaskType = TASK_NAME_TO_TASK_TYPE[task]
        data_root_dict: dict = TASK_NAME_TO_DATA_ROOT[task]

        train_data_roots: str = os.path.join(self.root_path, data_root_dict["train"])
        val_data_roots: str = os.path.join(self.root_path, data_root_dict["val"])
        test_data_roots: str = os.path.join(self.root_path, data_root_dict["test"])

        det_train_dataset_adapter = DetectionDatasetAdapter(
            task_type=task_type,
            train_data_roots=train_data_roots,
            val_data_roots=val_data_roots,
        )

        assert Subset.TRAINING in det_train_dataset_adapter.dataset
        assert Subset.VALIDATION in det_train_dataset_adapter.dataset

        det_train_dataset = det_train_dataset_adapter.get_otx_dataset()
        det_train_label_schema = det_train_dataset_adapter.get_label_schema()
        assert isinstance(det_train_dataset, DatasetEntity)
        assert isinstance(det_train_label_schema, LabelSchemaEntity)

        # In the test data, there is a empty_label image.
        # So, has_empty_label should be True
        has_empty_label = False
        for train_data in det_train_dataset:
            if isinstance(train_data.annotation_scene, NullAnnotationSceneEntity):
                has_empty_label = True
        assert has_empty_label is True

        det_test_dataset_adapter = DetectionDatasetAdapter(
            task_type=task_type,
            test_data_roots=test_data_roots,
        )

        assert Subset.TESTING in det_test_dataset_adapter.dataset
        assert isinstance(det_test_dataset_adapter.get_otx_dataset(), DatasetEntity)
        assert isinstance(det_test_dataset_adapter.get_label_schema(), LabelSchemaEntity)

    @e2e_pytest_unit
    def test_get_subset_data(self):
        class MockDatumDataset:
            def __init__(self, subsets):
                self._subsets = subsets
                self.eager = None

            def subsets(self):
                return self._subsets

        class MockDataset:
            def __init__(self, string):
                self.string = string

            def as_dataset(self):
                return self.string

        task = "detection"

        task_type: TaskType = TASK_NAME_TO_TASK_TYPE[task]
        data_root_dict: dict = TASK_NAME_TO_DATA_ROOT[task]

        train_data_roots: str = os.path.join(self.root_path, data_root_dict["train"])
        val_data_roots: str = os.path.join(self.root_path, data_root_dict["val"])

        det_train_dataset_adapter = DetectionDatasetAdapter(
            task_type=task_type,
            train_data_roots=train_data_roots,
            val_data_roots=val_data_roots,
        )

        dataset = MockDatumDataset(
            {
                "train": MockDataset("train"),
                "val": MockDataset("val"),
                "test": MockDataset("test"),
            }
        )
        assert det_train_dataset_adapter._get_subset_data("val", dataset) == "val"

        dataset = MockDatumDataset(
            {
                "train": MockDataset("train"),
                "validation": MockDataset("validation"),
                "test": MockDataset("test"),
            }
        )
        assert det_train_dataset_adapter._get_subset_data("val", dataset) == "validation"

        dataset = MockDatumDataset(
            {
                "train": MockDataset("train"),
                "trainval": MockDataset("trainval"),
                "val": MockDataset("val"),
            }
        )
        assert det_train_dataset_adapter._get_subset_data("val", dataset) == "val"

        dataset = MockDatumDataset(
            {
                "train2017": MockDataset("train2017"),
                "val2017": MockDataset("val2017"),
                "test2017": MockDataset("test2017"),
            }
        )
        assert det_train_dataset_adapter._get_subset_data("val", dataset) == "val2017"

    @e2e_pytest_unit
    def test_instance_segmentation(self):
        task = "instance_segmentation"

        task_type: TaskType = TASK_NAME_TO_TASK_TYPE[task]
        data_root_dict: dict = TASK_NAME_TO_DATA_ROOT[task]

        train_data_roots: str = os.path.join(self.root_path, data_root_dict["train"])
        val_data_roots: str = os.path.join(self.root_path, data_root_dict["val"])
        test_data_roots: str = os.path.join(self.root_path, data_root_dict["test"])

        instance_seg_train_dataset_adapter = DetectionDatasetAdapter(
            task_type=task_type,
            train_data_roots=train_data_roots,
            val_data_roots=val_data_roots,
        )

        assert Subset.TRAINING in instance_seg_train_dataset_adapter.dataset
        assert Subset.VALIDATION in instance_seg_train_dataset_adapter.dataset

        instance_seg_otx_train_data = instance_seg_train_dataset_adapter.get_otx_dataset()
        instance_seg_otx_train_label_schema = instance_seg_train_dataset_adapter.get_label_schema()
        assert isinstance(instance_seg_otx_train_data, DatasetEntity)
        assert isinstance(instance_seg_otx_train_label_schema, LabelSchemaEntity)

        # In the test data, there is a empty_label image.
        # So, has_empty_label should be True
        has_empty_label = False
        for train_data in instance_seg_otx_train_data:
            if isinstance(train_data.annotation_scene, NullAnnotationSceneEntity):
                has_empty_label = True
        assert has_empty_label is True

        instance_seg_test_dataset_adapter = DetectionDatasetAdapter(
            task_type=task_type,
            test_data_roots=test_data_roots,
        )

        assert Subset.TESTING in instance_seg_test_dataset_adapter.dataset
        assert isinstance(instance_seg_test_dataset_adapter.get_otx_dataset(), DatasetEntity)
        assert isinstance(instance_seg_test_dataset_adapter.get_label_schema(), LabelSchemaEntity)



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import openvino.runtime as ov
import pytest
import torch

from otx.algorithms.segmentation.adapters.mmseg.models.backbones import MMOVBackbone
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestMMOVBackbone:
    @pytest.fixture(autouse=True)
    def setup(self):
        param = ov.opset10.parameter([1, 3, 64, 64], ov.Type.f32, name="in")
        filter = ov.opset10.constant(np.random.normal(size=(1, 3, 64, 64)), ov.Type.f32)
        mul = ov.opset10.matmul(param, filter, False, False)
        result = ov.opset10.result(mul, name="out")
        ov_model = ov.Model([result], [param], "seg_backbone")

        self.model = MMOVBackbone(
            model_path_or_model=ov_model,
            outputs=["out"],
            remove_normalize=True,
            merge_bn=True,
            paired_bn=True,
        )

    @e2e_pytest_unit
    def test_init_weights(self):
        self.model.init_weights()

    @e2e_pytest_unit
    def test_forward(self):
        assert self.model.inputs == self.model._inputs
        assert self.model.outputs == self.model._outputs
        assert self.model.features == self.model._feature_dict
        assert self.model.input_shapes == self.model._input_shapes
        assert self.model.output_shapes == self.model._output_shapes

        data = {}
        for key, shape in self.model.input_shapes.items():
            shape = [1 if i == -1 else i for i in shape]
            data[key] = torch.randn(shape)
        self.model.train()
        self.model(list(data.values()), torch.tensor([0]))



import torch

from otx.algorithms.detection.adapters.mmdet.models.backbones import imgclsmob
from otx.algorithms.detection.adapters.mmdet.models.detectors.custom_single_stage_detector import (
    CustomSingleStageDetector,
)
from otx.algorithms.detection.adapters.mmdet.models.heads import custom_anchor_generator

__all__ = ["imgclsmob", "custom_anchor_generator"]

from mmdet.models.builder import build_detector

from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestCustomSingleStageDetector:
    @e2e_pytest_unit
    def test_custom_single_stage_detector_build(self):
        model_cfg = dict(
            type="CustomSingleStageDetector",
            backbone=dict(
                type="mmdet.mobilenetv2_w1", out_indices=(4, 5), frozen_stages=-1, norm_eval=False, pretrained=True
            ),
            neck=None,
            bbox_head=dict(
                type="CustomSSDHead",
                num_classes=80,
                in_channels=(int(96.0), int(320.0)),
                use_depthwise=True,
                norm_cfg=dict(type="BN"),
                act_cfg=dict(type="ReLU"),
                init_cfg=dict(type="Xavier", layer="Conv2d", distribution="uniform"),
                loss_balancing=False,
                anchor_generator=dict(
                    type="SSDAnchorGeneratorClustered",
                    strides=(16, 32),
                    reclustering_anchors=True,
                    widths=[
                        [
                            38.641007923271076,
                            92.49516032784699,
                            271.4234764938237,
                            141.53469410876247,
                        ],
                        [
                            206.04136086566515,
                            386.6542727907841,
                            716.9892752215089,
                            453.75609561761405,
                            788.4629155558277,
                        ],
                    ],
                    heights=[
                        [
                            48.9243877087132,
                            147.73088476194903,
                            158.23569788707474,
                            324.14510379107367,
                        ],
                        [
                            587.6216059488938,
                            381.60024152086544,
                            323.5988913027747,
                            702.7486097568518,
                            741.4865860938451,
                        ],
                    ],
                ),
                bbox_coder=dict(
                    type="DeltaXYWHBBoxCoder",
                    target_means=(0.0, 0.0, 0.0, 0.0),
                    target_stds=(0.1, 0.1, 0.2, 0.2),
                ),
            ),
        )

        model = build_detector(model_cfg)
        assert isinstance(model, CustomSingleStageDetector)

    @e2e_pytest_unit
    def test_custom_single_stage_detector_load_state_dict_pre_hook(self):
        chkpt_classes = ["person", "car"]
        model_classes = ["tree", "car", "person"]
        chkpt_dict = {
            "bbox_head.cls_convs.0.weight": torch.tensor(
                [
                    [1, 1, 1, 1],
                    [2, 2, 2, 2],
                    [0, 0, 0, 0],  # BG
                    [1, 1, 1, 1],
                    [2, 2, 2, 2],
                    [0, 0, 0, 0],  # BG
                    [1, 1, 1, 1],
                    [2, 2, 2, 2],
                    [0, 0, 0, 0],  # BG
                ]
            ),
            "bbox_head.cls_convs.0.bias": torch.tensor(
                [
                    [1],
                    [2],
                    [0],  # BG
                    [1],
                    [2],
                    [0],  # BG
                    [1],
                    [2],
                    [0],  # BG
                ]
            ),
        }
        model_dict = {
            "bbox_head.cls_convs.0.weight": torch.tensor(
                [
                    [3, 3, 3, 3],
                    [4, 4, 4, 4],
                    [5, 5, 5, 5],
                    [0, 0, 0, 0],  # BG
                    [3, 3, 3, 3],
                    [4, 4, 4, 4],
                    [5, 5, 5, 5],
                    [0, 0, 0, 0],  # BG
                    [3, 3, 3, 3],
                    [4, 4, 4, 4],
                    [5, 5, 5, 5],
                    [0, 0, 0, 0],  # BG
                ]
            ),
            "bbox_head.cls_convs.0.bias": torch.tensor(
                [
                    [3],
                    [4],
                    [5],
                    [0],  # BG
                    [3],
                    [4],
                    [5],
                    [0],  # BG
                    [3],
                    [4],
                    [5],
                    [0],  # BG
                ]
            ),
        }
        gt_dict = {
            "bbox_head.cls_convs.0.weight": torch.tensor(
                [
                    [3, 3, 3, 3],
                    [2, 2, 2, 2],
                    [1, 1, 1, 1],
                    [0, 0, 0, 0],  # BG
                    [3, 3, 3, 3],
                    [2, 2, 2, 2],
                    [1, 1, 1, 1],
                    [0, 0, 0, 0],  # BG
                    [3, 3, 3, 3],
                    [2, 2, 2, 2],
                    [1, 1, 1, 1],
                    [0, 0, 0, 0],  # BG
                ]
            ),
            "bbox_head.cls_convs.0.bias": torch.tensor(
                [
                    [3],
                    [2],
                    [1],
                    [0],  # BG
                    [3],
                    [2],
                    [1],
                    [0],  # BG
                    [3],
                    [2],
                    [1],
                    [0],  # BG
                ]
            ),
        }

        class Model:
            def state_dict(self):
                return model_dict

        CustomSingleStageDetector.load_state_dict_pre_hook(Model(), model_classes, chkpt_classes, chkpt_dict, "")
        for k, gt in gt_dict.items():
            assert (chkpt_dict[k] != gt).sum() == 0



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from otx.core.ov.graph import Graph
from otx.core.ov.graph.parsers.cls import cls_base_parser
from otx.core.ov.utils import load_ov_model
from tests.test_suite.e2e_test_system import e2e_pytest_unit


def get_graph(name="mobilenet-v2-pytorch"):
    model = load_ov_model(f"omz://{name}")
    return Graph.from_ov(model)


@e2e_pytest_unit
def test_cls_base_parser():
    graph = get_graph()
    assert {
        "inputs": ["data"],
        "outputs": ["/features/features#18/features#18#2/Clip"],
    } == cls_base_parser(graph, "backbone")
    assert {
        "inputs": ["/GlobalAveragePool"],
        "outputs": ["/Flatten"],
    } == cls_base_parser(graph, "neck")
    assert {
        "inputs": ["/classifier/classifier#1/Gemm/WithoutBiases"],
        "outputs": ["prob/sink_port_0"],
    } == cls_base_parser(graph, "head")



# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import string
from unittest.mock import patch

import numpy as np
import pytest
from torch.utils.data import DataLoader, Dataset

from otx.api.entities.annotation import AnnotationSceneEntity, AnnotationSceneKind
from otx.api.entities.dataset_item import DatasetItemEntity
from otx.api.entities.image import Image
from otx.core.data.caching import MemCacheHandlerSingleton
from otx.core.data.pipelines.load_image_from_otx_dataset import LoadImageFromOTXDataset


@pytest.fixture
def fxt_data_list():
    np.random.seed(3003)

    num_data = 10
    h = w = key_len = 16

    data_list = []
    for _ in range(num_data):
        data = np.random.randint(0, 256, size=[h, w, 3], dtype=np.uint8)
        key = "".join(
            [string.ascii_lowercase[i] for i in np.random.randint(0, len(string.ascii_lowercase), size=[key_len])]
        )
        data_list += [(key, data)]

    return data_list


@pytest.fixture
def fxt_caching_dataset_cls(fxt_data_list: list):
    class CachingDataset(Dataset):
        def __init__(self) -> None:
            super().__init__()
            self.d_items = [
                DatasetItemEntity(
                    media=Image(data=data),
                    annotation_scene=AnnotationSceneEntity(annotations=[], kind=AnnotationSceneKind.ANNOTATION),
                )
                for _, data in fxt_data_list
            ]
            self.load = LoadImageFromOTXDataset()

        def __len__(self):
            return len(self.d_items)

        def __getitem__(self, index):
            d_item = self.d_items[index]

            results = {
                "dataset_item": d_item,
                "height": d_item.media.numpy.shape[0],
                "width": d_item.media.numpy.shape[1],
                "index": index,
            }

            results = self.load(results)
            return results["img"]

    yield CachingDataset


def get_data_list_size(data_list):
    size = 0
    for _, data in data_list:
        size += data.size
    return size


class TestMemCacheHandler:
    @pytest.mark.parametrize("mode", ["singleprocessing", "multiprocessing"])
    def test_fully_caching(self, mode, fxt_data_list):
        mem_size = get_data_list_size(fxt_data_list)
        MemCacheHandlerSingleton.create(mode, mem_size)
        handler = MemCacheHandlerSingleton.get()

        for key, data in fxt_data_list:
            assert handler.put(key, data) > 0

        for key, data in fxt_data_list:
            get_data = handler.get(key)

            assert np.array_equal(get_data, data)

        # Fully cached
        assert len(handler) == len(fxt_data_list)

    @pytest.mark.parametrize("mode", ["singleprocessing", "multiprocessing"])
    def test_unfully_caching(self, mode, fxt_data_list):
        mem_size = get_data_list_size(fxt_data_list) // 2
        MemCacheHandlerSingleton.create(mode, mem_size)
        handler = MemCacheHandlerSingleton.get()

        for idx, (key, data) in enumerate(fxt_data_list):
            if idx < len(fxt_data_list) // 2:
                assert handler.put(key, data) > 0
            else:
                assert handler.put(key, data) is None

        for idx, (key, data) in enumerate(fxt_data_list):
            get_data = handler.get(key)

            if idx < len(fxt_data_list) // 2:
                assert np.array_equal(get_data, data)
            else:
                assert get_data is None

        # Unfully (half) cached
        assert len(handler) == len(fxt_data_list) // 2


class TestLoadImageFromFileWithCache:
    @pytest.mark.parametrize("mode", ["singleprocessing", "multiprocessing"])
    def test_combine_with_dataloader(self, mode, fxt_caching_dataset_cls, fxt_data_list):
        mem_size = get_data_list_size(fxt_data_list)
        MemCacheHandlerSingleton.create(mode, mem_size)

        dataset = fxt_caching_dataset_cls()

        with patch(
            "otx.core.data.pipelines.load_image_from_otx_dataset.get_image",
            side_effect=[data for _, data in fxt_data_list],
        ) as mock:
            for _ in DataLoader(dataset):
                continue

            # This initial round requires all data samples to be read from disk.
            assert mock.call_count == len(dataset)

        with patch(
            "otx.core.data.pipelines.load_image_from_otx_dataset.get_image",
            side_effect=[data for _, data in fxt_data_list],
        ) as mock:
            for _ in DataLoader(dataset):
                continue

            # The second round requires no read.
            assert mock.call_count == 0



"""Initial file for mmdetection dense heads."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .mmov_rpn_head import MMOVRPNHead
from .mmov_ssd_head import MMOVSSDHead
from .mmov_yolov3_head import MMOVYOLOV3Head

__all__ = ["MMOVRPNHead", "MMOVSSDHead", "MMOVYOLOV3Head"]



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import numpy as np
import openvino.runtime as ov
import pytest
import torch

from otx.algorithms.segmentation.adapters.mmseg.models import MMOVDecodeHead
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestMMOVBackbone:
    @pytest.fixture(autouse=True)
    def setup(self):

        params = []
        results = []

        param = ov.opset10.parameter([1, 24, 64, 64], ov.Type.f32, name="in")
        filter = ov.opset10.constant(np.random.normal(size=(1, 24, 64, 64)), ov.Type.f32)
        mul = ov.opset10.matmul(param, filter, False, False)
        result = ov.opset10.result(mul, name="extractor_out")
        params.append(param)
        results.append(result)

        filter = ov.opset10.constant(np.random.normal(size=(1, 24, 64, 64)), ov.Type.f32)
        mul = ov.opset10.matmul(param, filter, False, False)
        result = ov.opset10.result(mul, name="cls_seg_out")
        results.append(result)

        ov_model = ov.Model(results, params, "seg_head")

        self.model = MMOVDecodeHead(
            model_path_or_model=ov_model,
            inputs=dict(
                extractor="in",
                cls_seg="in",
            ),
            outputs=dict(
                extractor="extractor_out",
                cls_seg="cls_seg_out",
            ),
            in_channels=320,
            num_classes=24,
        )

    @e2e_pytest_unit
    def test_init_weights(self):
        self.model.init_weights()

    @e2e_pytest_unit
    def test_forward(self):

        data = {}
        input_shapes = self.model.conv_seg.input_shapes
        if getattr(self.model, "extractor", None):
            input_shapes = self.model.extractor.input_shapes

        for key, shape in input_shapes.items():
            shape = [1 if i == -1 else i for i in shape]
            data[key] = torch.randn(shape)

        output = self.model(list(data.values()))
        assert output.shape[1] == 24



"""Unit test for otx.algorithms.common.adapters.mmcv.hooks.eval_hook."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest
import torch
from mmcv.runner import BaseRunner
from torch.utils.data import DataLoader

from otx.algorithms.common.adapters.mmcv.hooks.eval_hook import (
    CustomEvalHook,
    DistCustomEvalHook,
    single_gpu_test,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class MockDataloader(DataLoader):
    """Mock class for pytorch dataloader."""

    class _MockDataset:
        def __init__(self, data) -> None:
            self.data = data

        def __len__(self) -> None:
            return len(self.data)

        def evaluate(self, results, *args, **kwargs):
            if len(results) == 0:
                return {"top-1": 0.5}
            return {"top-1": 0.7}

    def __init__(self) -> None:
        self.data = [{"img": torch.randn(1, 3, 224, 224)}]
        self.dataset = self._MockDataset(self.data)

    def __len__(self) -> int:
        return len(self.data)

    def __getitem__(self, idx: int) -> torch.Tensor:
        return self.data[idx]

    def __iter__(self):
        return iter(self.data)


class MockRunner(BaseRunner):
    """Mock class for BaseRunner."""

    class _ModuleWrapper:
        module = torch.nn.Module()

    class _MockBuffer:
        output = {}
        ready = False

        def clear(self):
            self.output = {}

    def __init__(self):
        self.model = torch.nn.Module()
        self.ema_model = self._ModuleWrapper()
        self._epoch = 9
        self._iter = 9
        self.log_buffer = self._MockBuffer()
        self.logger = None
        self.work_dir = "./tmp_dir/"

    def run(self):
        pass

    def save_checkpoint(self):
        pass

    def train(self):
        pass

    def val(self):
        pass


class TestCustomEvalHook:
    """Test class for CustomEvalHook."""

    @e2e_pytest_unit
    def test_init(self) -> None:
        """Test __init__ function."""

        hook = CustomEvalHook(metric="accuracy", dataloader=MockDataloader())
        assert hook.metric == "top-1"
        hook = CustomEvalHook(metric=["class_accuracy"], dataloader=MockDataloader())
        assert hook.metric == "accuracy"
        hook = CustomEvalHook(metric=["accuracy"], dataloader=MockDataloader())
        assert hook.metric == "top-1"

    @e2e_pytest_unit
    def test_do_evaluate(self, mocker) -> None:
        """Test do_evaluate function."""

        hook = CustomEvalHook(metric="accuracy", dataloader=MockDataloader())
        runner = MockRunner()
        mocker.patch("otx.algorithms.common.adapters.mmcv.hooks.eval_hook.single_gpu_test", return_value=[])
        mocker.patch.object(CustomEvalHook, "evaluate", return_value=True)
        hook._do_evaluate(runner, ema=False)
        hook.ema_eval_start_epoch = 3
        hook._do_evaluate(runner, ema=True)

    @e2e_pytest_unit
    def test_after_train_epoch(self, mocker) -> None:
        """Test after_train_epoch function."""

        hook = CustomEvalHook(metric="accuracy", dataloader=MockDataloader())
        runner = MockRunner()
        hook.by_epoch = True
        hook.interval = 4
        mocker.patch.object(CustomEvalHook, "_do_evaluate", return_value=True)
        hook.after_train_epoch(runner)
        hook.interval = 5
        hook.after_train_epoch(runner)

    @e2e_pytest_unit
    def test_after_train_iter(self, mocker) -> None:
        """Test after_train_iter function."""

        hook = CustomEvalHook(metric="accuracy", dataloader=MockDataloader())
        runner = MockRunner()
        hook.by_epoch = False
        hook.interval = 4
        mocker.patch.object(CustomEvalHook, "_do_evaluate", return_value=True)
        hook.after_train_iter(runner)
        hook.interval = 5
        hook.after_train_iter(runner)

    @e2e_pytest_unit
    def test_evaluate(self) -> None:
        """Test evaluate function."""

        hook = CustomEvalHook(metric="accuracy", dataloader=MockDataloader())
        runner = MockRunner()
        hook.evaluate(runner, results=[], results_ema=None)
        assert runner.log_buffer.output["top-1"] == 0.5
        assert runner.log_buffer.ready is True
        assert hook.best_score == 0.5
        assert runner.save_ckpt is True

        hook.evaluate(runner, results=[], results_ema=[0])
        assert runner.log_buffer.output["top-1_EMA"] == 0.7
        assert runner.save_ema_model is True


@e2e_pytest_unit
def test_single_gpu_test() -> None:
    """Test function for single_gpu_test."""

    class _MockModel(torch.nn.Module):
        def __init__(self):
            super().__init__()

        def forward(self, *args, **kwargs):
            return torch.Tensor([0])

    model = _MockModel()
    single_gpu_test(model, MockDataloader())


class TestDistCustomEvalHook:
    """Test class for DistCustomEvalHook."""

    @e2e_pytest_unit
    def test_do_evaluate(self, mocker) -> None:
        """Test _do_evaluate function."""

        mocker.patch("mmcls.apis.multi_gpu_test", return_value=True)
        mocker.patch.object(DistCustomEvalHook, "evaluate", return_value=True)
        runner = MockRunner()
        dataloader = MockDataloader()
        runner._rank = 0
        hook = DistCustomEvalHook(dataloader, metric="top-1")
        hook._do_evaluate(runner)

        with pytest.raises(TypeError):
            hook = DistCustomEvalHook(None)

    @e2e_pytest_unit
    def test_after_train_epoch(self, mocker) -> None:
        """Test after_train_epoch function."""

        dataloader = MockDataloader()
        hook = DistCustomEvalHook(dataloader, metric="top-1")
        mocker.patch.object(DistCustomEvalHook, "_do_evaluate", side_effect=RuntimeError("VALID ERROR"))
        hook.by_epoch = True
        hook.interval = 5
        runner = MockRunner()
        with pytest.raises(RuntimeError):
            hook.after_train_epoch(runner)

        hook.interval = 3
        hook.after_train_epoch(runner)

    @e2e_pytest_unit
    def test_after_train_iter(self, mocker) -> None:
        """Test after_train_iter function."""

        dataloader = MockDataloader()
        hook = DistCustomEvalHook(dataloader, metric="top-1")
        mocker.patch.object(DistCustomEvalHook, "_do_evaluate", side_effect=RuntimeError("VALID ERROR"))
        hook.by_epoch = False
        hook.interval = 5
        runner = MockRunner()
        with pytest.raises(RuntimeError):
            hook.after_train_iter(runner)

        hook.interval = 3
        hook.after_train_iter(runner)



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import pytest

from otx.algorithms.classification.adapters.mmcls.utils import patch_evaluation
from otx.algorithms.common.adapters.mmcv.utils.config_utils import MPAConfig
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@pytest.fixture
def otx_default_cls_config():
    config_name = "otx/algorithms/classification/configs/efficientnet_b0_cls_incr/model.py"
    conf = MPAConfig.fromfile(config_name)
    conf.checkpoint_config = {}
    conf.evaluation = {"metric": "None"}
    conf.data = {}
    return conf


@e2e_pytest_unit
def test_patch_evaluation(otx_default_cls_config) -> None:
    """Test patch_evaluation function.

    <Steps>
        1. Check eval metrics not empty
        2. Check the early stop metric is in eval metrics
        3. Check an exception is thrown for a frong task type
    """
    tasks = ["multilabel", "hierarchical", "normal"]
    for task in tasks:
        patch_evaluation(otx_default_cls_config, task)
        eval_metrics = otx_default_cls_config.evaluation.metric
        assert len(eval_metrics) > 0
        assert otx_default_cls_config.early_stop_metric in eval_metrics

    with pytest.raises(NotImplementedError):
        patch_evaluation(otx_default_cls_config, "None")



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import importlib

from mmcv.utils import Config

from otx.algorithms.common.adapters.mmdeploy.utils.mmdeploy import (
    is_mmdeploy_enabled,
    mmdeploy_init_model_helper,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit
from tests.unit.algorithms.common.adapters.mmdeploy.test_helpers import (
    create_config,
    create_model,
)


@e2e_pytest_unit
def test_is_mmdeploy_enabled():
    assert (importlib.util.find_spec("mmdeploy") is not None) == is_mmdeploy_enabled()


@e2e_pytest_unit
def test_mmdeploy_init_model_helper():
    from otx.algorithms.classification.adapters.mmcls.utils.builder import (
        build_classifier,
    )

    config = Config(
        {
            "model_cfg": create_config(),
            "device": "cpu",
        }
    )

    if importlib.util.find_spec("mmcls"):
        create_model("mmcls")
        model = mmdeploy_init_model_helper(config, model_builder=build_classifier)
        for i in model.parameters():
            assert not i.requires_grad



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from otx.core.patcher import Patcher
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class MockClass:
    def method(self, *args, **kwargs):
        pass

    @staticmethod
    def static_method(*args, **kwargs):
        pass

    @classmethod
    def class_method(cls, *args, **kwargs):
        pass


def mock_function(*args, **kwargs):
    pass


class Counter:
    def __init__(self):
        self._ctr = 0

    def inc(self):
        self._ctr += 1

    def __eq__(self, other):
        return self._ctr == other


class TestPatcher:
    @e2e_pytest_unit
    def test_patch(self):
        def dummy_wrapper(ctx, fn, *args, **kwargs):
            kwargs.get("ctr").inc()
            return fn(*args, **kwargs)

        def test_instance():
            patcher = Patcher()
            mock_class = MockClass()

            ctr = Counter()
            patcher.patch(mock_class.method, dummy_wrapper)
            patcher.patch(mock_class.method, dummy_wrapper)
            assert len(patcher._patched) == 1
            assert len(list(patcher._patched.values())[-1]) == 1
            mock_class.method(ctr=ctr)
            assert ctr == 1
            patcher.unpatch(mock_class.method)
            assert len(patcher._patched) == 0

            ctr = Counter()
            patcher.patch(mock_class.method, dummy_wrapper)
            patcher.patch(mock_class.method, dummy_wrapper, force=False)
            assert len(patcher._patched) == 1
            assert len(list(patcher._patched.values())[-1]) == 2
            mock_class.method(ctr=ctr)
            assert ctr == 2
            patcher.unpatch(mock_class.method, 1)
            assert len(list(patcher._patched.values())[-1]) == 1
            mock_class.method(ctr=ctr)
            assert ctr == 3

            ctr = Counter()
            patcher.patch(mock_class.method, dummy_wrapper, force=False)
            patcher.patch(mock_class.method, dummy_wrapper, force=False)
            patcher.patch(mock_class.method, dummy_wrapper, force=False)
            patcher.patch(mock_class.method, dummy_wrapper, force=False)
            assert len(list(patcher._patched.values())[-1]) == 5
            mock_class.method(ctr=ctr)
            assert ctr == 5
            patcher.unpatch(mock_class.method, 3)
            assert len(list(patcher._patched.values())[-1]) == 2
            mock_class.method(ctr=ctr)
            assert ctr == 7

            ctr = Counter()
            patcher.patch(mock_class.static_method, dummy_wrapper)
            patcher.patch(mock_class.static_method, dummy_wrapper)
            assert len(patcher._patched) == 2
            assert len(list(patcher._patched.values())[-1]) == 1
            mock_class.static_method(ctr=ctr)
            assert ctr == 1
            patcher.unpatch(mock_class.static_method)
            assert len(patcher._patched) == 1

            ctr = Counter()
            patcher.patch(mock_class.static_method, dummy_wrapper)
            patcher.patch(mock_class.static_method, dummy_wrapper, force=False)
            assert len(patcher._patched) == 2
            assert len(list(patcher._patched.values())[-1]) == 2
            mock_class.static_method(ctr=ctr)
            assert ctr == 2
            patcher.unpatch(mock_class.static_method, 1)
            assert len(list(patcher._patched.values())[-1]) == 1
            mock_class.static_method(ctr=ctr)
            assert ctr == 3

            ctr = Counter()
            patcher.patch(mock_class.static_method, dummy_wrapper, force=False)
            patcher.patch(mock_class.static_method, dummy_wrapper, force=False)
            patcher.patch(mock_class.static_method, dummy_wrapper, force=False)
            patcher.patch(mock_class.static_method, dummy_wrapper, force=False)
            assert len(list(patcher._patched.values())[-1]) == 5
            mock_class.static_method(ctr=ctr)
            assert ctr == 5
            patcher.unpatch(mock_class.static_method, 3)
            assert len(list(patcher._patched.values())[-1]) == 2
            mock_class.static_method(ctr=ctr)
            assert ctr == 7

            ctr = Counter()
            patcher.patch(mock_class.class_method, dummy_wrapper)
            patcher.patch(mock_class.class_method, dummy_wrapper)
            assert len(patcher._patched) == 3
            assert len(list(patcher._patched.values())[-1]) == 1
            mock_class.class_method(ctr=ctr)
            assert ctr == 1
            patcher.unpatch(mock_class.class_method)
            assert len(patcher._patched) == 2

            ctr = Counter()
            patcher.patch(mock_class.class_method, dummy_wrapper)
            patcher.patch(mock_class.class_method, dummy_wrapper, force=False)
            assert len(patcher._patched) == 3
            assert len(list(patcher._patched.values())[-1]) == 2
            mock_class.class_method(ctr=ctr)
            assert ctr == 2
            patcher.unpatch(mock_class.class_method, 1)
            assert len(list(patcher._patched.values())[-1]) == 1
            mock_class.class_method(ctr=ctr)
            assert ctr == 3

            ctr = Counter()
            patcher.patch(mock_class.class_method, dummy_wrapper, force=False)
            patcher.patch(mock_class.class_method, dummy_wrapper, force=False)
            patcher.patch(mock_class.class_method, dummy_wrapper, force=False)
            patcher.patch(mock_class.class_method, dummy_wrapper, force=False)
            assert len(list(patcher._patched.values())[-1]) == 5
            mock_class.class_method(ctr=ctr)
            assert ctr == 5
            patcher.unpatch(mock_class.class_method, 3)
            assert len(list(patcher._patched.values())[-1]) == 2
            mock_class.class_method(ctr=ctr)
            assert ctr == 7

            patcher.unpatch()
            assert len(patcher._patched) == 0

        def test_class():
            patcher = Patcher()
            mock_class = MockClass()

            ctr = Counter()
            patcher.patch(MockClass.method, dummy_wrapper)
            patcher.patch(MockClass.method, dummy_wrapper)
            assert len(patcher._patched) == 1
            assert len(list(patcher._patched.values())[-1]) == 1
            mock_class.method(ctr=ctr)
            assert ctr == 1
            patcher.unpatch(MockClass.method)
            assert len(patcher._patched) == 0

            ctr = Counter()
            patcher.patch(MockClass.method, dummy_wrapper)
            patcher.patch(MockClass.method, dummy_wrapper, force=False)
            assert len(patcher._patched) == 1
            assert len(list(patcher._patched.values())[-1]) == 2
            mock_class.method(ctr=ctr)
            assert ctr == 2
            patcher.unpatch(MockClass.method, 1)
            assert len(list(patcher._patched.values())[-1]) == 1
            mock_class.method(ctr=ctr)
            assert ctr == 3

            ctr = Counter()
            patcher.patch(MockClass.method, dummy_wrapper, force=False)
            patcher.patch(MockClass.method, dummy_wrapper, force=False)
            patcher.patch(MockClass.method, dummy_wrapper, force=False)
            patcher.patch(MockClass.method, dummy_wrapper, force=False)
            assert len(list(patcher._patched.values())[-1]) == 5
            mock_class.method(ctr=ctr)
            assert ctr == 5
            patcher.unpatch(MockClass.method, 3)
            assert len(list(patcher._patched.values())[-1]) == 2
            mock_class.method(ctr=ctr)
            assert ctr == 7

            ctr = Counter()
            patcher.patch(MockClass.static_method, dummy_wrapper)
            patcher.patch(MockClass.static_method, dummy_wrapper)
            assert len(patcher._patched) == 2
            assert len(list(patcher._patched.values())[-1]) == 1
            mock_class.static_method(ctr=ctr)
            assert ctr == 1
            patcher.unpatch(MockClass.static_method)
            assert len(patcher._patched) == 1

            ctr = Counter()
            patcher.patch(MockClass.static_method, dummy_wrapper)
            patcher.patch(MockClass.static_method, dummy_wrapper, force=False)
            assert len(patcher._patched) == 2
            assert len(list(patcher._patched.values())[-1]) == 2
            mock_class.static_method(ctr=ctr)
            assert ctr == 2
            patcher.unpatch(MockClass.static_method, 1)
            assert len(list(patcher._patched.values())[-1]) == 1
            mock_class.static_method(ctr=ctr)
            assert ctr == 3

            ctr = Counter()
            patcher.patch(MockClass.static_method, dummy_wrapper, force=False)
            patcher.patch(MockClass.static_method, dummy_wrapper, force=False)
            patcher.patch(MockClass.static_method, dummy_wrapper, force=False)
            patcher.patch(MockClass.static_method, dummy_wrapper, force=False)
            assert len(list(patcher._patched.values())[-1]) == 5
            mock_class.static_method(ctr=ctr)
            assert ctr == 5
            patcher.unpatch(MockClass.static_method, 3)
            assert len(list(patcher._patched.values())[-1]) == 2
            mock_class.static_method(ctr=ctr)
            assert ctr == 7

            ctr = Counter()
            patcher.patch(MockClass.class_method, dummy_wrapper)
            patcher.patch(MockClass.class_method, dummy_wrapper)
            assert len(patcher._patched) == 3
            assert len(list(patcher._patched.values())[-1]) == 1
            mock_class.class_method(ctr=ctr)
            assert ctr == 1
            patcher.unpatch(MockClass.class_method)
            assert len(patcher._patched) == 2

            ctr = Counter()
            patcher.patch(MockClass.class_method, dummy_wrapper)
            patcher.patch(MockClass.class_method, dummy_wrapper, force=False)
            assert len(patcher._patched) == 3
            assert len(list(patcher._patched.values())[-1]) == 2
            mock_class.class_method(ctr=ctr)
            assert ctr == 2
            patcher.unpatch(MockClass.class_method, 1)
            assert len(list(patcher._patched.values())[-1]) == 1
            mock_class.class_method(ctr=ctr)
            assert ctr == 3

            ctr = Counter()
            patcher.patch(MockClass.class_method, dummy_wrapper, force=False)
            patcher.patch(MockClass.class_method, dummy_wrapper, force=False)
            patcher.patch(MockClass.class_method, dummy_wrapper, force=False)
            patcher.patch(MockClass.class_method, dummy_wrapper, force=False)
            assert len(list(patcher._patched.values())[-1]) == 5
            mock_class.class_method(ctr=ctr)
            assert ctr == 5
            patcher.unpatch(MockClass.class_method, 3)
            assert len(list(patcher._patched.values())[-1]) == 2
            mock_class.class_method(ctr=ctr)
            assert ctr == 7

            patcher.unpatch()
            assert len(patcher._patched) == 0

        def test_module():
            patcher = Patcher()

            ctr = Counter()
            patcher.patch(
                "tests.unit.core.test_core_patcher.mock_function",
                dummy_wrapper,
            )
            patcher.patch(
                "tests.unit.core.test_core_patcher.mock_function",
                dummy_wrapper,
            )
            assert len(patcher._patched) == 1
            assert len(list(patcher._patched.values())[-1]) == 1
            from tests.unit.core.test_core_patcher import mock_function

            mock_function(ctr=ctr)
            assert ctr == 1
            patcher.unpatch("tests.unit.core.test_core_patcher.mock_function")
            assert len(patcher._patched) == 0

            ctr = Counter()
            patcher.patch(
                "tests.unit.core.test_core_patcher.mock_function",
                dummy_wrapper,
            )
            patcher.patch(
                "tests.unit.core.test_core_patcher.mock_function",
                dummy_wrapper,
                force=False,
            )
            assert len(list(patcher._patched.values())[-1]) == 2
            from tests.unit.core.test_core_patcher import mock_function

            mock_function(ctr=ctr)
            assert ctr == 2
            patcher.unpatch(
                "tests.unit.core.test_core_patcher.mock_function",
                1,
            )
            assert len(list(patcher._patched.values())[-1]) == 1
            from tests.unit.core.test_core_patcher import mock_function

            mock_function(ctr=ctr)
            assert ctr == 3

            ctr = Counter()
            patcher.patch(
                "tests.unit.core.test_core_patcher.mock_function",
                dummy_wrapper,
                force=False,
            )
            patcher.patch(
                "tests.unit.core.test_core_patcher.mock_function",
                dummy_wrapper,
                force=False,
            )
            patcher.patch(
                "tests.unit.core.test_core_patcher.mock_function",
                dummy_wrapper,
                force=False,
            )
            patcher.patch(
                "tests.unit.core.test_core_patcher.mock_function",
                dummy_wrapper,
                force=False,
            )
            assert len(list(patcher._patched.values())[-1]) == 5
            from tests.unit.core.test_core_patcher import mock_function

            mock_function(ctr=ctr)
            assert ctr == 5
            patcher.unpatch(
                "tests.unit.core.test_core_patcher.mock_function",
                3,
            )
            assert len(list(patcher._patched.values())[-1]) == 2
            from tests.unit.core.test_core_patcher import mock_function

            mock_function(ctr=ctr)
            assert ctr == 7

            patcher.unpatch()
            assert len(patcher._patched) == 0

        test_instance()
        test_class()
        test_module()

    @e2e_pytest_unit
    def test_import_obj(self):
        patcher = Patcher()
        assert (Patcher, "patch") == patcher.import_obj("otx.core.patcher.Patcher.patch")
        assert (Patcher, "patch") == patcher.import_obj(Patcher.patch)



# Copyright (C) 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import abc

import pytest

from otx.api.entities.media import IMedia2DEntity, IMediaEntity
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestIMediaEntity:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_i_media_entity(self):
        """
        <b>Description:</b>
        To test IMediaEntity class

        <b>Input data:</b>
        Instance of IMediaEntity class

        <b>Expected results:</b>
        1. Test instance is instance of class IMediaEntity


        <b>Steps</b>
        1. Create IMediaEntity

        """
        test_inst = IMediaEntity()
        assert isinstance(test_inst, IMediaEntity)


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestIMedia2DEntity:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_i_media_2d_entity(self):
        """
        <b>Description:</b>
        To test IMedia2DEntity abstract class

        <b>Input data:</b>
        Instance of IMedia2DEntity abstract class

        <b>Expected results:</b>
        1. TypeError is raised
        2. Expected method numbers and names


        <b>Steps</b>
        1. Create IMedia2DEntity
        2. Check abstract methods

        """
        with pytest.raises(TypeError):
            IMedia2DEntity()

        assert type(IMedia2DEntity) is abc.ABCMeta
        abc_methods = IMedia2DEntity.__abstractmethods__
        assert len(abc_methods) == 4
        assert "width" in abc_methods
        assert "roi_numpy" in abc_methods
        assert "numpy" in abc_methods
        assert "height" in abc_methods



"""ROI Extractor list for mmdetection adapters."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

from .single_level_roi_extractor import SingleRoIExtractor

__all__ = ["SingleRoIExtractor"]



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import tempfile
from pathlib import Path
from time import sleep

import pytest

from otx.api.usecases.exportable_code.streamer import (
    CameraStreamer,
    DirStreamer,
    ImageStreamer,
    OpenError,
    ThreadedStreamer,
    VideoStreamer,
    get_streamer,
)
from tests.test_helpers import (
    generate_random_image_folder,
    generate_random_single_image,
    generate_random_single_video,
    generate_random_video_folder,
)
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestStreamer:
    @staticmethod
    def assert_streamer_element(streamer):
        for element in streamer:
            assert element.shape == (360, 480, 3)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_image_streamer_with_single_image(self):
        """
        <b>Description:</b>
        Test that ImageStreamer works correctly with a single image as input

        <b>Input data:</b>
        Random image file

        <b>Expected results:</b>
        Test passes if ImageStreamer returns a single image with the correct size

        <b>Steps</b>
        1. Create ImageStreamer
        2. Request image from streamer
        """
        with generate_random_single_image(height=360, width=480) as path:
            streamer = ImageStreamer(path)
            self.assert_streamer_element(streamer)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_dir_streamer_with_folder(self):
        """
        <b>Description:</b>
        Test that DirStreamer works correctly with a folder of images as input

        <b>Input data:</b>
        Folder with 10 random images

        <b>Expected results:</b>
        Test passes if DirStreamer returns ten images with the correct size

        <b>Steps</b>
        1. Create DirStreamer
        2. Request images from streamer
        """
        with generate_random_image_folder(height=360, width=480) as path:
            streamer = DirStreamer(path)
            self.assert_streamer_element(streamer)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_video_streamer_with_single_video(self):
        """
        <b>Description:</b>
        Test that VideoStreamer works correctly with a single video as input

        <b>Input data:</b>
        Random Video file

        <b>Expected results:</b>
        Test passes if VideoStreamer can read the Video file frame by frame

        <b>Steps</b>
        1. Create VideoStreamer
        2. Request frame from VideoStreamer
        """
        with generate_random_single_video(height=360, width=480) as path:
            streamer = VideoStreamer(path)
            self.assert_streamer_element(streamer)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_video_streamer_with_loop_flag(self):
        """
        <b>Description:</b>
        Test that VideoStreamer works correctly with a loop flag

        <b>Input data:</b>
        Random Video file

        <b>Expected results:</b>
        Test passes if VideoStreamer returns frames with the correct amount of dimensions
        after the end of the video

        <b>Steps</b>
        1. Create VideoStreamer
        2. Request frames from streamer
        """
        with generate_random_single_video(height=360, width=480, number_of_frames=100) as path:
            streamer = VideoStreamer(path, loop=True)

            for index, frame in enumerate(streamer):
                assert frame.shape[-1] == 3
                if index > 200:
                    break

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_video_streamer_with_single_image(self):
        """
        <b>Description:</b>
        Test that VideoStreamer works correctly with a single image as input

        <b>Input data:</b>
        Random image file

        <b>Expected results:</b>
        Test passes if VideoStreamer can read the single frame

        <b>Steps</b>
        1. Create VideoStreamer
        2. Request frame from VideoStreamer
        """
        with generate_random_single_video(height=360, width=480) as path:
            streamer = VideoStreamer(path)
            self.assert_streamer_element(streamer)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_invalid_inputs_to_get_streamer(self):
        """
        <b>Description:</b>
        Test that get_streamer does not allow invalid inputs

        <b>Input data:</b>
        Invalid file
        Empty directory
        Folder with random videos
        Name of file that does not exist

        <b>Expected results:</b>
        Test passes if get_streamer raises a ValueError

        <b>Steps</b>
        1. Create invalid input file with .bin extension
        2. Attempt to call get_streamer with .bin file
        3. Attempt to call get_streamer with empty directory
        4. Attempt to call get_streamer with video folder, this raises because it is not supported
        5. Attempt to call get_streamer with a string that does not point to a file
        """
        with tempfile.TemporaryDirectory() as temp_dir:
            invalid_file = Path(temp_dir) / "not_valid.bin"
            invalid_file.touch()

            with pytest.raises(Exception) as context:
                get_streamer(str(invalid_file))

        the_exception = context  # .exception
        assert "Can't open" in str(the_exception), str(the_exception)

        with tempfile.TemporaryDirectory() as empty_dir:
            with pytest.raises(Exception):
                get_streamer(empty_dir)

        with generate_random_video_folder() as path:
            with pytest.raises(Exception):
                get_streamer(path)

        with pytest.raises(Exception) as context:
            get_streamer("not_a_file")

        the_exception = context  # .exception
        assert "Can't find" in str(the_exception), str(the_exception)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_valid_inputs_to_get_streamer(self):
        """
        <b>Description:</b>
        Test that get_streamer return the correct Streamer class for each input

        <b>Input data:</b>
        A Video
        An Image
        Folder with images

        <b>Expected results:</b>
        Test passes if each call to get_streamer return the correct Streamer instance

        <b>Steps</b>
        1. Call get_streamer with a video file
        2. Call get_streamer with an image file
        3. Call get_streamer with a folder of images
        4. Call get_streamer with a camera index
        5. Call get_streamer with the threaded argument
        """
        with generate_random_single_video() as path:
            streamer = get_streamer(path)
            assert isinstance(streamer, VideoStreamer)

        with generate_random_single_image() as path:
            streamer = get_streamer(path)
            assert isinstance(streamer, ImageStreamer)

        with generate_random_image_folder() as path:
            streamer = get_streamer(path)
            assert isinstance(streamer, DirStreamer)

        streamer = get_streamer(0)
        assert isinstance(streamer, CameraStreamer)

        streamer = get_streamer(input_stream=0, threaded=True)
        assert isinstance(streamer, ThreadedStreamer)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_video_file_fails_on_image_streamer(self):
        """
        <b>Description:</b>
        Test that ImageStreamer raises an exception if a video is passed

        <b>Input data:</b>
        Random Video file

        <b>Expected results:</b>
        Test passes if a OpenError is raised

        <b>Steps</b>
        1. Attempt to create ImageStreamer
        """
        with generate_random_single_video() as path:
            with pytest.raises(OpenError):
                ImageStreamer(path)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_camera_streamer(self):
        """
        <b>Description:</b>
        Check that CameraStreamer works correctly

        <b>Input data:</b>
        CameraStreamer

        <b>Expected results:</b>
        Test passes if CameraStreamer can read 10 frames from the camera

        <b>Steps</b>
        1. Create camera streamer
        2. Retrieve frames from camera streamer
        """
        streamer = get_streamer()
        n = 10

        for frame in streamer:
            assert frame.shape[-1] == 3
            n -= 1
            if n == 0:
                break

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    @pytest.mark.timeout(10)
    @pytest.mark.xfail(reason="CVS-102619")
    def test_threaded_streamer(self):
        """
        <b>Description:</b>
        Check that ThreadedStreamer works correctly

        <b>Input data:</b>
        Folder with images

        <b>Expected results:</b>
        Test passes if ThreadedStreamer reads 3 frames from the folder

        <b>Steps</b>
        1. Create ThreadedStreamer
        2. Retrieve frames from ThreadedStreamer
        """
        with generate_random_image_folder() as path:
            streamer = get_streamer(path, threaded=True)
            frame_count = 0

            for frame in streamer:
                assert frame.shape[-1] == 3
                frame_count += 1

                if frame_count == 3:
                    break

            assert frame_count == 3

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_threaded_streamer_timeout(self):
        """
        <b>Description:</b>
        Check that ThreadedStreamer works correctly even if the main thread is very slow

        <b>Input data:</b>
        Folder with images

        <b>Expected results:</b>
        Test passes if ThreadedStreamer returns 5 images

        <b>Steps</b>
        1. Create ThreadedStreamer
        2. Retrieve frames from camera streamer and suspend each thread in between frames
        """
        with generate_random_image_folder() as path:
            streamer = get_streamer(path, threaded=True)

            streamer.buffer_size = 2
            frame_count = 0

            for frame in streamer:
                assert frame.shape[-1] == 3
                sleep(1)
                frame_count += 1
                if frame_count == 5:
                    break

            assert frame_count == 5



"""Quantization preset Enums for post training optimization."""

# Copyright (C) 2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

from otx.api.configuration import ConfigurableEnum


class POTQuantizationPreset(ConfigurableEnum):
    """This Enum represents the quantization preset for post training optimization."""

    PERFORMANCE = "Performance"
    MIXED = "Mixed"


class StorageCacheScheme(ConfigurableEnum):
    """This Enum represents the storage scheme for Datumaro arrow format."""

    NONE = "NONE"
    AS_IS = "AS-IS"
    JPEG_75 = "JPEG/75"
    JPEG_95 = "JPEG/95"
    PNG = "PNG"
    TIFF = "TIFF"



"""Unit tests for OTX Cli"""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import sys

import pytest

from otx.cli.tools import cli
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestCli:
    @e2e_pytest_unit
    def test_cli(self):
        backup_argv = sys.argv
        # invalid inputs
        sys.argv = ["otx"]
        with pytest.raises(SystemExit) as e:
            cli.main()
        assert e.type == SystemExit, f"{e}"
        sys.argv = ["otx", None]
        with pytest.raises(SystemExit) as e:
            cli.main()
        assert e.type == SystemExit, f"{e}"
        sys.argv = ["otx", 0]
        with pytest.raises(SystemExit) as e:
            cli.main()
        assert e.type == SystemExit, f"{e}"
        sys.argv = ["otx", ""]
        with pytest.raises(SystemExit) as e:
            cli.main()
        assert e.type == SystemExit, f"{e}"
        # not a string value
        sys.argv = ["otx", -1]
        with pytest.raises(Exception) as e:
            cli.main()
        assert e.type == TypeError, f"{e}"
        sys.argv = ["otx", b"\x00"]
        with pytest.raises(Exception) as e:
            cli.main()
        assert e.type == TypeError, f"{e}"
        sys.argv = backup_argv



"""Utils for common OTX algorithms."""

# Copyright (C) 2022 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import copy
import glob
import os
import os.path as osp
import platform
import shutil
import sys
import tempfile
import warnings
from collections.abc import Mapping
from importlib import import_module
from typing import Any, Callable, Dict, List, Tuple, Union

import torch
from mmcv import Config, ConfigDict
from mmcv.utils.config import BASE_KEY, DEPRECATION_KEY
from mmcv.utils.misc import import_modules_from_strings
from mmcv.utils.path import check_file_exist

from otx.algorithms.common.utils.logger import get_logger
from otx.api.entities.datasets import DatasetEntity

from ._config_utils_get_configs_by_keys import get_configs_by_keys
from ._config_utils_get_configs_by_pairs import get_configs_by_pairs

logger = get_logger()


# TODO: refactor Config
class MPAConfig(Config):
    """A class that extends the base `Config` class, adds additional functionality for loading configuration files."""

    @staticmethod
    def _file2dict(
        filename, use_predefined_variables=True
    ):  # pylint: disable=too-many-locals, too-many-branches, too-many-statements
        """Static method that loads the configuration file and returns a dictionary of its contents.

        :param filename: str, the path of the configuration file to be loaded.
        :param use_predefined_variables: bool, a flag indicating whether to substitute predefined variables in the
                                          configuration file.
        :return: tuple of dictionary and string. Returns a dictionary containing the contents of the configuration file
                 and a string representation of the configuration file.
        :raises: IOError if the file type is not supported.
        """
        filename = osp.abspath(osp.expanduser(filename))
        check_file_exist(filename)
        extender = osp.splitext(filename)[1]
        if extender not in [".py", ".json", ".yaml", ".yml"]:
            raise IOError("Only py/yml/yaml/json type are supported now!")

        with tempfile.TemporaryDirectory() as temp_config_dir:
            with tempfile.NamedTemporaryFile(dir=temp_config_dir, suffix=extender) as temp_config_file:
                if platform.system() == "Windows":
                    temp_config_file.close()
                temp_config_name = osp.basename(temp_config_file.name)
                # Substitute predefined variables
                if use_predefined_variables:
                    Config._substitute_predefined_vars(filename, temp_config_file.name)
                else:
                    shutil.copyfile(filename, temp_config_file.name)
                # Substitute base variables from placeholders to strings
                base_var_dict = Config._pre_substitute_base_vars(temp_config_file.name, temp_config_file.name)
                if filename.endswith(".py"):
                    temp_module_name = osp.splitext(temp_config_name)[0]
                    sys.path.insert(0, temp_config_dir)
                    Config._validate_py_syntax(filename)
                    mod = import_module(temp_module_name)
                    sys.path.pop(0)
                    cfg_dict = {name: value for name, value in mod.__dict__.items() if not name.startswith("__")}
                    # delete imported module
                    del sys.modules[temp_module_name]
                elif filename.endswith((".yml", ".yaml", ".json")):
                    import mmcv

                    cfg_dict = mmcv.load(temp_config_file.name)

        # check deprecation information
        if DEPRECATION_KEY in cfg_dict:
            deprecation_info = cfg_dict.pop(DEPRECATION_KEY)
            warning_msg = f"The config file {filename} will be deprecated " "in the future."
            if "expected" in deprecation_info:
                warning_msg += f' Please use {deprecation_info["expected"]} ' "instead."
            if "reference" in deprecation_info:
                warning_msg += " More information can be found at " f'{deprecation_info["reference"]}'
            warnings.warn(warning_msg)

        cfg_text = filename + "\n"
        with open(filename, "r", encoding="utf-8") as f:
            # Setting encoding explicitly to resolve coding issue on windows
            cfg_text += f.read()

        if BASE_KEY in cfg_dict:
            cfg_dir = osp.dirname(filename)
            base_filename = cfg_dict.pop(BASE_KEY)
            base_filename = base_filename if isinstance(base_filename, list) else [base_filename]

            cfg_dict_list = []
            cfg_text_list = []
            for f in base_filename:
                _cfg_dict, _cfg_text = MPAConfig._file2dict(osp.join(cfg_dir, f))
                cfg_dict_list.append(_cfg_dict)
                cfg_text_list.append(_cfg_text)

            base_cfg_dict = dict()
            # for c in cfg_dict_list:
            #     duplicate_keys = base_cfg_dict.keys() & c.keys()
            #     if len(duplicate_keys) > 0:
            #         raise KeyError('Duplicate key is not allowed among bases. '
            #                        f'Duplicate keys: {duplicate_keys}')
            #     base_cfg_dict.update(c)
            for c in cfg_dict_list:
                if len(base_cfg_dict.keys() & c.keys()) > 0:
                    # raise KeyError(f'Duplicate key is not allowed among bases [{base_cfg_dict.keys() & c.keys()}]')
                    logger.warning(f"Duplicate key is detected among bases [{base_cfg_dict.keys() & c.keys()}]")
                    logger.debug(f"base = {base_cfg_dict}, cfg = {c}")
                    base_cfg_dict = Config._merge_a_into_b(base_cfg_dict, c)
                    logger.debug(f"merged dict = {base_cfg_dict}")
                else:
                    base_cfg_dict.update(c)

            # Subtitute base variables from strings to their actual values
            cfg_dict = Config._substitute_base_vars(cfg_dict, base_var_dict, base_cfg_dict)

            base_cfg_dict = Config._merge_a_into_b(cfg_dict, base_cfg_dict)
            cfg_dict = base_cfg_dict

            # merge cfg_text
            cfg_text_list.append(cfg_text)
            cfg_text = "\n".join(cfg_text_list)

        return cfg_dict, cfg_text

    @staticmethod
    def fromfile(filename, use_predefined_variables=True, import_custom_modules=True):
        """Static method that loads a configuration file and returns an instance of `Config` class.

        :param filename: str, the path of the configuration file to be loaded.
        :param use_predefined_variables: bool, a flag indicating whether to substitute predefined variables in the
                                          configuration file.
        :param import_custom_modules: bool, a flag indicating whether to import custom modules.
        :return: Config object, an instance of `Config` class containing the contents of the configuration file.
        """
        cfg_dict, cfg_text = MPAConfig._file2dict(filename, use_predefined_variables)
        if import_custom_modules and cfg_dict.get("custom_imports", None):
            import_modules_from_strings(**cfg_dict["custom_imports"])
        return Config(cfg_dict, cfg_text=cfg_text, filename=filename)


def copy_config(cfg):
    """A function that creates a deep copy of the input configuration object.

    :param cfg: Config object, an instance of `Config` class to be copied.
    :return: Config object, a deep copy of the input configuration object.
    :raises: ValueError if the input object is not an instance of `Config` class.
    """
    if not isinstance(cfg, Config):
        raise ValueError(f"cannot copy this instance {type(cfg)}")
    # new_cfg = copy.deepcopy(cfg)
    # new_cfg._cfg_dict = copy.deepcopy(cfg._cfg_dict)
    # new_cfg.filename = cfg.filename
    import pickle

    data = pickle.dumps(cfg)
    return pickle.loads(data)


def update_or_add_custom_hook(cfg: Config, hook_cfg: ConfigDict):
    """Update hook cfg if same type is in custom_hook or append it."""
    custom_hooks = cfg.get("custom_hooks", [])
    custom_hooks_updated = False
    for custom_hook in custom_hooks:
        if custom_hook["type"] == hook_cfg["type"]:
            custom_hook.update(hook_cfg)
            custom_hooks_updated = True
            break
    if not custom_hooks_updated:
        custom_hooks.append(hook_cfg)
    cfg["custom_hooks"] = custom_hooks


def remove_custom_hook(cfg: Config, hook_type: str):
    """Remove hook cfg if hook_type is in custom_hook."""
    custom_hooks = cfg.get("custom_hooks", [])
    if len(custom_hooks) > 0:
        idx_to_del = None
        for i, custom_hook in enumerate(custom_hooks):
            if custom_hook["type"] == hook_type:
                idx_to_del = i
                break
        if idx_to_del is not None:
            del custom_hooks[idx_to_del]


def recursively_update_cfg(
    cfg: Union[Config, dict],
    criterion: Callable[[Any, Any], bool],
    update_dict: Any,
):
    """A function that recursively updates the input dictionary or `Config` object with a new dictionary.

    :param cfg: Union[Config, dict], an input dictionary or `Config` object to be updated.
    :param criterion: Callable[[Any, Any], bool], a function that determines whether to update a key-value pair based on
                      a criterion. The function takes two arguments: key and value, and returns a boolean.
    :param update_dict: Any, a dictionary to be used for updating the input dictionary.
    :return: None
    """
    for key, val in list(cfg.items()):
        if isinstance(val, dict):
            recursively_update_cfg(val, criterion, update_dict)
        if criterion(key, val):
            cfg.update(update_dict)


def add_custom_hook_if_not_exists(cfg: Config, hook_cfg: ConfigDict):
    """A function that adds a custom hook to the input `Config` object if it doesn't already exist.

    :param cfg: Config object, an instance of `Config` class to which the custom hook will be added.
    :param hook_cfg: ConfigDict object, an instance of `ConfigDict` class representing the custom hook to be added.
    :return: None
    """
    custom_hooks = cfg.get("custom_hooks", [])
    found = False
    for hook in custom_hooks:
        if hook["type"] == hook_cfg["type"]:
            found = True
            break
    if not found:
        custom_hooks.append(hook_cfg)
        cfg["custom_hooks"] = custom_hooks


def remove_from_config(config: Union[Config, ConfigDict], key: str):
    """Update & Remove configs."""
    if key in config:
        if isinstance(config, Config):
            del config._cfg_dict[key]  # pylint: disable=protected-access
        elif isinstance(config, ConfigDict):
            del config[key]
        else:
            raise ValueError(f"Unknown config type {type(config)}")


def remove_from_configs_by_type(configs: List[ConfigDict], type_name: str):
    """Update & remove by type."""
    indices = []
    for i, config in enumerate(configs):
        type_name_ = config.get("type", None)
        if type_name_ == type_name:
            indices.append(i)
    for i in reversed(indices):
        configs.pop(i)


def update_config(
    config: Union[Config, ConfigDict],
    pairs: Dict[Tuple[Any, ...], Any],
):
    """Update configs by path as a key and value as a target."""
    for path, value in pairs.items():
        path_ = list(reversed(path))
        ptr = config
        key = None
        while path_:
            key = path_.pop()
            if isinstance(ptr, (Config, Mapping)):
                if key not in ptr:
                    ptr[key] = ConfigDict()
            elif isinstance(ptr, (list, tuple)):
                assert isinstance(key, int), f"{key} of {path} must be int for ({type(ptr)}: {ptr})"
                assert len(ptr) < key, f"{key} of {path} exceeds {len(ptr)}"
            if len(path_) == 0:
                ptr[key] = value
            ptr = ptr[key]


def get_dataset_configs(config: Union[Config, ConfigDict], subset: str) -> List[ConfigDict]:
    """A function that retrieves 'datasets' configurations from the input `Config` object or `ConfigDict` object.

    :param config: Union[Config, ConfigDict], an instance of `Config` class or `ConfigDict` class containing the
                   configurations.
    :param subset: str, a string representing the subset for which the 'datasets' configuration is required.
    :return: List[ConfigDict], a list of 'datasets' configuration dictionaries.
    """
    if config.data.get(subset, None) is None:
        return []
    data_cfg = config.data[subset]
    data_cfgs = get_configs_by_keys(data_cfg, ["dataset", "datasets"])
    return data_cfgs if data_cfgs else [data_cfg]


def prepare_for_testing(config: Union[Config, ConfigDict], dataset: DatasetEntity) -> Config:
    """Prepare configs for testing phase."""
    config = copy.deepcopy(config)
    # FIXME. Should working directories be modified here?
    config.data.test.otx_dataset = dataset
    return config


def is_epoch_based_runner(runner_config: ConfigDict):
    """Check Epoch based or Iter based runner."""
    return "Epoch" in runner_config.type


def config_from_string(config_string: str) -> Config:
    """Generate an mmcv config dict object from a string.

    :param config_string: string to parse
    :return config: configuration object
    """
    with tempfile.NamedTemporaryFile("w", suffix=".py") as temp_file:
        temp_file.write(config_string)
        temp_file.flush()
        return Config.fromfile(temp_file.name)


def patch_default_config(config: Config):
    """Patch default config."""
    if "runner" not in config:
        config.runner = ConfigDict({"type": "EpochBasedRunner"})
    if "log_config" not in config:
        config.log_config = ConfigDict()
    if "evaluation" not in config:
        config.evaluation = ConfigDict()
    if "checkpoint_config" not in config:
        config.checkpoint_config = ConfigDict({"type": "CheckpointHook", "interval": 1})


def patch_data_pipeline(config: Config, data_pipeline: str = ""):
    """Replace data pipeline to data_pipeline.py if it exist."""
    if os.path.isfile(data_pipeline):
        data_pipeline_cfg = Config.fromfile(data_pipeline)
        config.merge_from_dict(data_pipeline_cfg)
    else:
        raise FileNotFoundError(f"data_pipeline: {data_pipeline} not founded")


def patch_color_conversion(config: Config):
    """Patch color conversion."""
    assert "data" in config

    for cfg in get_configs_by_pairs(config.data, dict(type="Normalize")):
        to_rgb = False
        if "to_rgb" in cfg:
            to_rgb = cfg.to_rgb
        cfg.to_rgb = not bool(to_rgb)


def patch_runner(config: Config):
    """Patch runner."""
    assert "runner" in config

    # Check that there is no conflict in specification of number of training epochs.
    # Move global definition of epochs inside runner config.
    if "total_epochs" in config:
        if is_epoch_based_runner(config.runner):
            if config.runner.max_epochs != config.total_epochs:
                logger.warning("Conflicting declaration of training epochs number.")
            config.runner.max_epochs = config.total_epochs
        else:
            logger.warning(f"Total number of epochs set for an iteration based runner {config.runner.type}.")
        remove_from_config(config, "total_epochs")

    # Change runner's type.
    if config.runner.type != "AccuracyAwareRunner":
        if is_epoch_based_runner(config.runner) and config.runner.type != "EpochRunnerWithCancel":
            logger.info(f"Replacing runner from {config.runner.type} to EpochRunnerWithCancel.")
            config.runner.type = "EpochRunnerWithCancel"
        elif not is_epoch_based_runner(config.runner) and config.runner.type != "IterBasedRunnerWithCancel":
            logger.info(f"Replacing runner from {config.runner.type} to IterBasedRunnerWithCancel.")
            config.runner.type = "IterBasedRunnerWithCancel"


def patch_fp16(config: Config):
    """Remove FP16 config if running on CPU device and revert to FP32.

    Please refer https://github.com/pytorch/pytorch/issues/23377
    """
    if not torch.cuda.is_available() and "fp16" in config:
        logger.info("Revert FP16 to FP32 on CPU device")
        if isinstance(config, Config):
            del config._cfg_dict["fp16"]  # pylint: disable=protected-access
        elif isinstance(config, ConfigDict):
            del config["fp16"]


def patch_adaptive_interval_training(config: Config):
    """Update adaptive interval settings for OTX training.

    This function can be removed by adding custom hook cfg into recipe.py directly.
    """
    # default adaptive hook for evaluating before and after training
    add_custom_hook_if_not_exists(
        config,
        ConfigDict(
            type="AdaptiveTrainSchedulingHook",
            enable_adaptive_interval_hook=False,
            enable_eval_before_run=True,
        ),
    )
    # Add/remove adaptive interval hook
    if config.get("use_adaptive_interval", False):
        update_or_add_custom_hook(
            config,
            ConfigDict(
                {
                    "type": "AdaptiveTrainSchedulingHook",
                    "max_interval": 5,
                    "enable_adaptive_interval_hook": True,
                    "enable_eval_before_run": True,
                    **config.pop("adaptive_validation_interval", {}),
                }
            ),
        )
    else:
        config.pop("adaptive_validation_interval", None)


def patch_early_stopping(config: Config):
    """Update early stop settings for OTX training.

    This function can be removed by adding custom hook cfg into recipe.py directly.
    """
    if "early_stop" in config:
        remove_custom_hook(config, "EarlyStoppingHook")
        early_stop = config.get("early_stop", False)
        if early_stop:
            early_stop_hook = ConfigDict(
                type="LazyEarlyStoppingHook",
                start=early_stop.start,
                patience=early_stop.patience,
                iteration_patience=early_stop.iteration_patience,
                interval=1,
                metric=config.early_stop_metric,
                priority=75,
            )
            update_or_add_custom_hook(config, early_stop_hook)
        else:
            remove_custom_hook(config, "LazyEarlyStoppingHook")

    # make sure model to be in a training mode even after model is evaluated (mmcv bug)
    update_or_add_custom_hook(
        config,
        ConfigDict(type="ForceTrainModeHook", priority="LOWEST"),
    )


def patch_persistent_workers(config: Config):
    """If num_workers is 0, persistent_workers must be False."""
    data_cfg = config.data
    for subset in ["train", "val", "test", "unlabeled"]:
        if subset not in data_cfg:
            continue
        dataloader_cfg = data_cfg.get(f"{subset}_dataloader", ConfigDict())
        workers_per_gpu = dataloader_cfg.get(
            "workers_per_gpu",
            data_cfg.get("workers_per_gpu", 0),
        )
        if workers_per_gpu == 0:
            dataloader_cfg["persistent_workers"] = False
            data_cfg[f"{subset}_dataloader"] = dataloader_cfg


def patch_from_hyperparams(config: Config, hyperparams):
    """Patch config parameters from hyperparams."""
    params = hyperparams.learning_parameters
    warmup_iters = int(params.learning_rate_warmup_iters)
    lr_config = (
        ConfigDict(warmup_iters=warmup_iters)
        if warmup_iters > 0
        else ConfigDict(warmup_iters=warmup_iters, warmup=None)
    )

    if params.enable_early_stopping and config.get("evaluation", None):
        early_stop = ConfigDict(
            start=int(params.early_stop_start),
            patience=int(params.early_stop_patience),
            iteration_patience=int(params.early_stop_iteration_patience),
        )
    else:
        early_stop = False

    runner = ConfigDict(max_epochs=int(params.num_iters))
    if config.get("runner", None) and config.runner.get("type").startswith("IterBasedRunner"):
        runner = ConfigDict(max_iters=int(params.num_iters))

    hparams = ConfigDict(
        optimizer=ConfigDict(lr=params.learning_rate),
        lr_config=lr_config,
        early_stop=early_stop,
        data=ConfigDict(
            samples_per_gpu=int(params.batch_size),
            workers_per_gpu=int(params.num_workers),
        ),
        runner=runner,
    )

    hparams["use_adaptive_interval"] = hyperparams.learning_parameters.use_adaptive_interval
    config.merge_from_dict(hparams)


def align_data_config_with_recipe(data_config: ConfigDict, config: Union[Config, ConfigDict]):
    """Align data_cfg with recipe_cfg."""
    # we assumed config has 'otx_dataset' and 'labels' key in it
    # by 'patch_datasets' function

    data_config = data_config.data
    config = config.data
    for subset in data_config.keys():
        subset_config = data_config.get(subset, {})
        for key in list(subset_config.keys()):
            found_config = get_configs_by_keys(config.get(subset), key, return_path=True)
            assert len(found_config) == 1
            value = subset_config.pop(key)
            path = list(found_config.keys())[0]
            update_config(subset_config, {path: value})


DEFAULT_META_KEYS = (
    "filename",
    "ori_filename",
    "ori_shape",
    "img_shape",
    "pad_shape",
    "scale_factor",
    "flip",
    "flip_direction",
    "img_norm_cfg",
)


def get_meta_keys(pipeline_step, add_meta_keys: List[str] = []):
    """Update meta_keys for ignore_labels."""
    meta_keys = list(pipeline_step.get("meta_keys", DEFAULT_META_KEYS))
    meta_keys.append("ignored_labels")
    meta_keys += add_meta_keys
    pipeline_step["meta_keys"] = set(meta_keys)
    return pipeline_step


def prepare_work_dir(config: Union[Config, ConfigDict]) -> str:
    """Prepare configs of working directory."""
    base_work_dir = config.work_dir
    checkpoint_dirs = glob.glob(os.path.join(base_work_dir, "checkpoints_round_*"))
    train_round_checkpoint_dir = os.path.join(base_work_dir, f"checkpoints_round_{len(checkpoint_dirs)}")
    os.makedirs(train_round_checkpoint_dir)
    config.work_dir = train_round_checkpoint_dir
    if "meta" not in config.runner:
        config.runner.meta = ConfigDict()
    config.runner.meta.exp_name = f"train_round_{len(checkpoint_dirs)}"
    return train_round_checkpoint_dir


def get_data_cfg(config: Union[Config, ConfigDict], subset: str = "train") -> Config:
    """Return dataset configs."""
    data_cfg = config.data[subset]
    while "dataset" in data_cfg:
        data_cfg = data_cfg.dataset
    return data_cfg



# Copyright (C) 2021-2023 Intel Corporation
# SPDX-Lggggcense-Identifier: Apache-2.0
#

import openvino.runtime as ov
import pytest
import torch

from otx.core.ov.ops.movements import get_torch_padding
from otx.core.ov.ops.utils import get_dynamic_shape
from tests.test_suite.e2e_test_system import e2e_pytest_unit


@e2e_pytest_unit
def test_get_dynamic_shape():
    op_ov = ov.opset10.parameter([-1, 256, 256, 3], ov.Type.f32)
    output = get_dynamic_shape(op_ov)
    assert output == [-1, 256, 256, 3]


@e2e_pytest_unit
def test_get_torch_padding():
    input = torch.randn(4, 3, 64, 64)

    output = get_torch_padding([5, 5], [5, 5], "valid", input.shape[2:], [3, 3], [1, 1])
    assert output == 0

    output = get_torch_padding([5, 5], [5, 5], "same_upper", input.shape[2:], [3, 3], [1, 1])
    output = output(input)
    assert output.shape == (4, 3, 66, 66)

    output = get_torch_padding([5, 5], [5, 5], "explicit", input.shape[2:], [3, 3], [1, 1])
    output = output(input)
    assert output.shape == (4, 3, 74, 74)

    with pytest.raises(NotImplementedError):
        get_torch_padding([5, 5], [5, 5], "error", input.shape[2:], [3, 3], [1, 1])



import torch
from mmcv.utils import ConfigDict
from mmdet.models.builder import build_detector

from otx.algorithms.detection.adapters.mmdet.models.detectors.custom_vfnet_detector import (
    CustomVFNet,
)
from tests.test_suite.e2e_test_system import e2e_pytest_unit


class TestCustomVFNet:
    @e2e_pytest_unit
    def test_custom_vfnet_build(self):
        model_cfg = ConfigDict(
            type="CustomVFNet",
            backbone=dict(
                type="ResNet",
                depth=50,
                num_stages=4,
                out_indices=(0, 1, 2, 3),
                frozen_stages=1,
                norm_cfg=dict(type="SyncBN", requires_grad=True),
                norm_eval=True,
                style="pytorch",
            ),
            neck=dict(
                type="FPN",
                in_channels=[256, 512, 1024, 2048],
                out_channels=256,
                start_level=1,
                add_extra_convs="on_output",
                num_outs=5,
                relu_before_extra_convs=True,
            ),
            bbox_head=dict(
                type="VFNetHead",
                num_classes=2,
                in_channels=256,
                stacked_convs=3,
                feat_channels=256,
                strides=[8, 16, 32, 64, 128],
                center_sampling=False,
                dcn_on_last_conv=False,
                use_atss=True,
                use_vfl=True,
                loss_cls=dict(
                    type="VarifocalLoss", use_sigmoid=True, alpha=0.75, gamma=2.0, iou_weighted=True, loss_weight=1.0
                ),
                loss_bbox=dict(type="GIoULoss", loss_weight=1.5),
                loss_bbox_refine=dict(type="GIoULoss", loss_weight=2.0),
            ),
            train_cfg=dict(assigner=dict(type="ATSSAssigner", topk=9), allowed_border=-1, pos_weight=-1, debug=False),
            test_cfg=dict(
                nms_pre=1000, min_bbox_size=0, score_thr=0.01, nms=dict(type="nms", iou_threshold=0.5), max_per_img=100
            ),
            task_adapt=dict(
                src_classes=["person", "car"],
                dst_classes=["tree", "car", "person"],
            ),
        )

        model = build_detector(model_cfg)
        assert isinstance(model, CustomVFNet)

    @e2e_pytest_unit
    def test_custom_vfnet_load_state_dict_pre_hook(self):
        chkpt_classes = ["person", "car"]
        model_classes = ["tree", "car", "person"]
        chkpt_dict = {
            "bbox_head.vfnet_cls.weight": torch.tensor(
                [
                    [1, 1, 1, 1],
                    [2, 2, 2, 2],
                ]
            ),
            "bbox_head.vfnet_cls.bias": torch.tensor(
                [
                    [1],
                    [2],
                ]
            ),
        }
        model_dict = {
            "bbox_head.vfnet_cls.weight": torch.tensor(
                [
                    [3, 3, 3, 3],
                    [4, 4, 4, 4],
                    [5, 5, 5, 5],
                ]
            ),
            "bbox_head.vfnet_cls.bias": torch.tensor(
                [
                    [3],
                    [4],
                    [5],
                ]
            ),
        }
        gt_dict = {
            "bbox_head.vfnet_cls.weight": torch.tensor(
                [
                    [3, 3, 3, 3],
                    [2, 2, 2, 2],
                    [1, 1, 1, 1],
                ]
            ),
            "bbox_head.vfnet_cls.bias": torch.tensor(
                [
                    [3],
                    [2],
                    [1],
                ]
            ),
        }

        class Model:
            def state_dict(self):
                return model_dict

        CustomVFNet.load_state_dict_pre_hook(Model(), model_classes, chkpt_classes, chkpt_dict, "")
        for k, gt in gt_dict.items():
            assert (chkpt_dict[k] != gt).sum() == 0



# Copyright (C) 2020-2021 Intel Corporation
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions
# and limitations under the License.

import itertools
import warnings
from datetime import datetime

import pytest

from otx.api.entities.shapes.ellipse import Ellipse
from otx.api.entities.shapes.polygon import Point, Polygon
from otx.api.entities.shapes.rectangle import Rectangle
from otx.api.entities.shapes.shape import (
    GeometryException,
    Shape,
    ShapeEntity,
    ShapeType,
)
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestShapeType:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_shapetype(self):
        """
        <b>Description:</b>
        Check ShapeType class length returns expected value

        <b>Expected results:</b>
        Test passes if ShapeType enum class length equal expected value
        """
        assert len(ShapeType) == 3
        assert ShapeType.ELLIPSE.value == 1
        assert ShapeType.RECTANGLE.value == 2
        assert ShapeType.POLYGON.value == 3


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestShapeEntity:
    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_shape_entity_not_implemented_methods(self):
        """
        <b>Description:</b>
        Check not implemented methods of ShapeEntity class

        <b>Expected results:</b>
        Test passes if NotImplementedError exception raises when using not implemented methods on ShapeEntity instance
        """
        rectangle_entity = Rectangle(x1=0.2, y1=0.2, x2=0.6, y2=0.7)
        ellipse_entity = Ellipse(x1=0.4, y1=0.1, x2=0.9, y2=0.8)
        polygon_entity = Polygon(
            [
                Point(0.3, 0.4),
                Point(0.3, 0.7),
                Point(0.5, 0.75),
                Point(0.8, 0.7),
                Point(0.8, 0.4),
            ]
        )
        for shape in [rectangle_entity, ellipse_entity, polygon_entity]:
            with pytest.raises(NotImplementedError):
                ShapeEntity.get_area(shape)
            with pytest.raises(NotImplementedError):
                ShapeEntity.intersects(shape, shape)
            with pytest.raises(NotImplementedError):
                ShapeEntity.contains_center(shape, shape)
            with pytest.raises(NotImplementedError):
                ShapeEntity.normalize_wrt_roi_shape(shape, rectangle_entity)
            with pytest.raises(NotImplementedError):
                ShapeEntity.denormalize_wrt_roi_shape(shape, rectangle_entity)
            with pytest.raises(NotImplementedError):
                ShapeEntity._as_shapely_polygon(shape)


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestShape:
    @staticmethod
    def fully_covering_rectangle() -> Rectangle:
        return Rectangle.generate_full_box()

    @staticmethod
    def fully_covering_ellipse() -> Ellipse:
        return Ellipse(x1=0.0, y1=0.0, x2=1.0, y2=1.0)

    @staticmethod
    def fully_covering_polygon() -> Polygon:
        return Polygon(
            [
                Point(0.0, 0.1),
                Point(0.0, 0.9),
                Point(0.5, 1.0),
                Point(1.0, 0.9),
                Point(1.0, 0.0),
                Point(0.0, 0.1),
            ]
        )

    def rectangle(self) -> Rectangle:
        return Rectangle(x1=0.2, y1=0.2, x2=0.6, y2=0.7)

    def ellipse(self) -> Ellipse:
        return Ellipse(x1=0.4, y1=0.1, x2=0.9, y2=0.8)

    def polygon(self) -> Polygon:
        return Polygon(
            [
                Point(0.3, 0.4),
                Point(0.3, 0.7),
                Point(0.5, 0.75),
                Point(0.8, 0.7),
                Point(0.8, 0.4),
                Point(0.3, 0.4),
            ],
        )

    @staticmethod
    def not_inscribed_rectangle() -> Rectangle:
        return Rectangle(x1=0.0, y1=0.0, x2=0.01, y2=0.01)

    @staticmethod
    def not_inscribed_ellipse() -> Ellipse:
        return Ellipse(x1=0.0, y1=0.0, x2=0.01, y2=0.01)

    @staticmethod
    def not_inscribed_polygon() -> Polygon:
        return Polygon(
            [
                Point(0.0, 0.0),
                Point(0.0, 0.01),
                Point(0.01, 0.02),
                Point(0.02, 0.01),
                Point(0.02, 0.0),
                Point(0.0, 0.0),
            ]
        )

    @staticmethod
    def base_self_intersect_polygon() -> Polygon:
        return Polygon(
            [
                Point(0.3, 0.3),
                Point(0.4, 0.3),
                Point(0.3, 0.3),
                Point(0.3, 0.2),
                Point(0.3, 1),
                Point(0.2, 0.2),
            ]
        )

    @staticmethod
    def other_self_intersect_polygon() -> Polygon:
        return Polygon(
            [
                Point(0.3, 0.2),
                Point(0.2, 0.3),
                Point(0.3, 0.1),
                Point(0.3, 0.2),
                Point(0, 0.2),
                Point(0, 4),
            ]
        )

    @staticmethod
    def lower_side_intersect_shapes() -> list:
        return [
            Rectangle(x1=0.2, y1=0.1, x2=0.5, y2=0.4),
            Polygon(
                [
                    Point(0.35, 0.1),
                    Point(0.2, 0.2),
                    Point(0.2, 0.4),
                    Point(0.5, 0.4),
                    Point(0.5, 0.2),
                    Point(0.35, 0.1),
                ]
            ),
        ]

    @staticmethod
    def upper_side_intersect_shapes() -> list:
        return [
            Rectangle(x1=0.2, y1=0.4, x2=0.5, y2=0.7),
            Polygon(
                [
                    Point(0.35, 0.7),
                    Point(0.2, 0.6),
                    Point(0.2, 0.4),
                    Point(0.5, 0.4),
                    Point(0.5, 0.6),
                    Point(0.35, 0.7),
                ]
            ),
        ]

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_shape_get_area_method(self):
        """
        <b>Description:</b>
        Check get_area not implemented method of Shape class

        <b>Expected results:</b>
        Test passes if NotImplementedError exception raised when using get_area method on Shape instance
        """
        for shape in [self.rectangle(), self.ellipse(), self.polygon()]:
            with pytest.raises(NotImplementedError):
                Shape.get_area(shape)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_shape_magic_methods(self):
        """
        <b>Description:</b>
        Check __repr__ and __hash__ methods for Shape class instance

        <b>Expected results:</b>
        Test passes if __repr__ and __hash__ method return expected values

        <b>Steps</b>
        1. Check that __repr__ method returns expected value
        2. Check that __hash__ method returns expected value
        """
        test_rectangle = Rectangle(0.0, 0.0, 1.0, 1.0, modification_date=datetime(year=2021, month=11, day=23))
        assert Shape.__repr__(test_rectangle) == "Shape with modification date:('2021-11-23 00:00:00')"
        assert Shape.__hash__(test_rectangle) == hash(str(test_rectangle))

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_shape_intersects(self):
        """
        <b>Description:</b>
        Check Shape intersects method for Rectangle, Ellipse and Polygon objects

        <b>Expected results:</b>
        Test passes if intersects method returns expected values

        <b>Steps</b>
        1. Check intersects method when Shapes intersect completely
        2. Check intersects method when Shapes intersect in several points
        3. Check intersects method when Shapes intersect by one side
        4. Check intersects method when Shapes not intersect
        5. Check GeometryException exception raised with incorrect parameters for intersect method
        """
        inscribed_shapes_list = [self.rectangle(), self.ellipse(), self.polygon()]
        # Check when Shapes intersect fully
        for full_element in [
            self.fully_covering_rectangle(),
            self.fully_covering_ellipse(),
            self.fully_covering_polygon(),
        ]:
            for inscribed in inscribed_shapes_list:
                assert full_element.intersects(inscribed)
                assert inscribed.intersects(full_element)
        # Check when Shapes intersect in several points
        for shape, other_shape in list(itertools.combinations(inscribed_shapes_list, 2)):
            assert shape.intersects(other_shape)
            assert other_shape.intersects(shape)
        # Check when Shapes intersect by one side
        for upper_shape in self.upper_side_intersect_shapes():
            for lower_shape in self.lower_side_intersect_shapes():
                assert lower_shape.intersects(upper_shape)
        # Check when Shapes not intersect
        for shape in inscribed_shapes_list:
            for not_inscribed_shape in (
                self.not_inscribed_rectangle(),
                self.not_inscribed_ellipse(),
                self.not_inscribed_polygon(),
            ):
                assert not shape.intersects(not_inscribed_shape)
                assert not not_inscribed_shape.intersects(shape)
        # Checking GeometryException exception raised
        with pytest.raises(GeometryException):
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", "Polygon coordinates")
                self.base_self_intersect_polygon().intersects(self.other_self_intersect_polygon())

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_shape_contains_center(self):
        """
        <b>Description:</b>
        Check Shape contains_center method for Rectangle, Ellipse and Polygon objects

        <b>Expected results:</b>
        Test passes if contains_center method returns expected values

        <b>Steps</b>
        1. Check contains_center method when a Polygon, Rectangle and Ellipse fall within a Rectangle
        2. Check contains_center method when a Polygon, Rectangle and Ellipse fall outside a Rectangle
        """
        rectangle_full = self.fully_covering_rectangle()
        shapes_inside = [self.polygon(), self.ellipse(), self.rectangle()]

        rectangle_part = self.rectangle()
        shapes_outside = [
            self.not_inscribed_polygon(),
            self.not_inscribed_ellipse(),
            self.not_inscribed_rectangle(),
        ]

        for shape_inside in shapes_inside:
            assert rectangle_full.contains_center(shape_inside)
        for shape_outside in shapes_outside:
            assert not rectangle_part.contains_center(shape_outside)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_validate_coordinates(self):
        """
        <b>Description:</b>
        Check Shape validate_coordinates method for Rectangle, Ellipse and Polygon objects

        <b>Expected results:</b>
        Test passes if validate_coordinates method returns expected values

        <b>Steps</b>
        1. Check validate_coordinates method for Shapes with 0.0<=x,y<=1.0
        2. Check validate_coordinates method for Shapes with x<0.0
        3. Check validate_coordinates method for Shapes with x>1.0
        4. Check validate_coordinates method for Shapes with y<0.0
        5. Check validate_coordinates method for Shapes with y>1.0
        6. Check validate_coordinates method for Shapes with x,y<0.0
        7. Check validate_coordinates method for Shapes with x,y>1.0
        8. Check validate_coordinates method for Shapes with x>1.0, y<0.0
        9. Check validate_coordinates method for Shapes with x<1.0, y>1.0
        """
        with warnings.catch_warnings():
            warnings.filterwarnings("ignore", r".* coordinates")
            for shape in [self.rectangle(), self.ellipse(), self.polygon()]:
                assert shape._validate_coordinates(x=0.0, y=0.0)
                assert shape._validate_coordinates(x=1.0, y=1.0)
                assert shape._validate_coordinates(x=0.2, y=0.3)
                assert not shape._validate_coordinates(x=-0.1, y=0.0)
                assert not shape._validate_coordinates(x=1.1, y=1.0)
                assert not shape._validate_coordinates(x=0.2, y=-0.3)
                assert not shape._validate_coordinates(x=0.2, y=1.3)
                assert not shape._validate_coordinates(x=-0.1, y=-0.2)
                assert not shape._validate_coordinates(x=1.1, y=1.2)
                assert not shape._validate_coordinates(x=1.2, y=-0.3)
                assert not shape._validate_coordinates(x=-1.2, y=1.3)



"""Pipeline element that loads an image from a OTX Dataset on the fly."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

from tempfile import TemporaryDirectory
from typing import Any, Dict, Tuple

import numpy as np

from otx.algorithms.common.utils.data import get_image

from ..caching import MemCacheHandlerError, MemCacheHandlerSingleton

_CACHE_DIR = TemporaryDirectory(prefix="img-cache-")  # pylint: disable=consider-using-with

# TODO: refactoring to common modules
# TODO: refactoring to Sphinx style.


class LoadImageFromOTXDataset:
    """Pipeline element that loads an image from a OTX Dataset on the fly.

    Can do conversion to float 32 if needed.
    Expected entries in the 'results' dict that should be passed to this pipeline element are:
        results['dataset_item']: dataset_item from which to load the image
        results['dataset_id']: id of the dataset to which the item belongs
        results['index']: index of the item in the dataset

    :param to_float32: optional bool, True to convert images to fp32. defaults to False
    """

    def __init__(self, to_float32: bool = False):
        self.to_float32 = to_float32
        try:
            self.mem_cache_handler = MemCacheHandlerSingleton.get()
        except MemCacheHandlerError:
            # Create a null handler
            MemCacheHandlerSingleton.create(mode="null", mem_size=0)
            self.mem_cache_handler = MemCacheHandlerSingleton.get()

    @staticmethod
    def _get_unique_key(results: Dict[str, Any]) -> Tuple:
        # TODO: We should improve it by assigning an unique id to DatasetItemEntity.
        # This is because there is a case which
        # d_item.media.path is None, but d_item.media.data is not None
        d_item = results["dataset_item"]
        return d_item.media.path, d_item.roi.id

    def __call__(self, results: Dict[str, Any]):
        """Callback function of LoadImageFromOTXDataset."""
        key = self._get_unique_key(results)

        img = self.mem_cache_handler.get(key)

        if img is None:
            # Get image (possibly from cache)
            img = get_image(results, _CACHE_DIR.name, to_float32=False)
            self.mem_cache_handler.put(key, img)

        if self.to_float32:
            img = img.astype(np.float32)
        shape = img.shape

        if img.shape[0] != results["height"]:
            results["height"] = img.shape[0]

        if img.shape[1] != results["width"]:
            results["width"] = img.shape[1]

        filename = f"Dataset item index {results['index']}"
        results["filename"] = filename
        results["ori_filename"] = filename
        results["img"] = img
        results["img_shape"] = shape
        results["ori_shape"] = shape
        # Set initial values for default meta_keys
        results["pad_shape"] = shape
        num_channels = 1 if len(shape) < 3 else shape[2]
        results["img_norm_cfg"] = dict(
            mean=np.zeros(num_channels, dtype=np.float32),
            std=np.ones(num_channels, dtype=np.float32),
            to_rgb=False,
        )
        results["img_fields"] = ["img"]
        results["entity_id"] = results.get("entity_id")
        results["label_id"] = results.get("label_id")

        return results



""" Runs tests selectively depending on changed files. """

import os
import sys
from subprocess import run

from tests.test_suite.run_test_command import collect_env_vars

ALGO_ROOT_DIR = "external"
ALGO_DIRS = [
    os.path.join(ALGO_ROOT_DIR, d) for d in os.listdir(ALGO_ROOT_DIR) if os.path.isdir(os.path.join(ALGO_ROOT_DIR, d))
]
IMPORTANT_DIRS = [
    "ote_cli/",
    "ote_sdk/",
    "tests/",
]

wd = sys.argv[1]


def what_to_test():
    """
    Returns a dict containing information whether it is needed
    to run tests for particular algorithm.
    """

    print(f"{sys.argv=}")
    run_algo_tests = {d: True for d in ALGO_DIRS}
    if len(sys.argv) > 2:
        run_algo_tests = {d: False for d in ALGO_DIRS}
        changed_files = sys.argv[2:]
        print(f"{changed_files=}")

        for changed_file in changed_files:
            if any(changed_file.startswith(d) for d in IMPORTANT_DIRS):
                run_algo_tests = {d: True for d in ALGO_DIRS}
                break

            for d in ALGO_DIRS:
                if changed_file.startswith(d):
                    run_algo_tests[d] = True

    for k, v in run_algo_tests.items():
        print("run", k, v)

    return run_algo_tests


def test(run_algo_tests):
    """
    Runs tests for algorithms and other stuff (misc).
    """

    passed = {}
    success = True
    command = ["pytest", os.path.join("tests", "ote_cli", "misc"), "-v"]
    try:
        res = run(command, env=collect_env_vars(wd), check=True).returncode == 0
    except:  # noqa: E722
        res = False
    passed["misc"] = res
    success *= res
    for algo_dir in ALGO_DIRS:
        if run_algo_tests[algo_dir]:
            command = ["pytest", os.path.join(algo_dir, "tests", "ote_cli"), "-v", "-rxXs", "--durations=10"]
            try:
                res = run(command, env=collect_env_vars(wd), check=True).returncode == 0
            except:  # noqa: E722
                res = False
            passed[algo_dir] = res
            success *= res

    for k, v in passed.items():
        res = "PASSED" if v else "FAILED"
        print(f"Tests for {k} {res}")

    sys.exit(1 - success)


test(what_to_test())



# Copyright (C) 2021-2022 Intel Corporation
# SPDX-License-Identifier: Apache-2.0
#

import datetime

import numpy as np
import pytest

from otx.api.entities.annotation import (
    Annotation,
    AnnotationSceneEntity,
    AnnotationSceneKind,
)
from otx.api.entities.color import Color
from otx.api.entities.id import ID
from otx.api.entities.label import Domain, LabelEntity
from otx.api.entities.result_media import ResultMediaEntity
from otx.api.entities.scored_label import ScoredLabel
from otx.api.entities.shapes.rectangle import Rectangle
from tests.unit.api.constants.components import OtxSdkComponent
from tests.unit.api.constants.requirements import Requirements

RANDOM_IMAGE = np.random.randint(low=0, high=255, size=(32, 64, 3))


@pytest.mark.components(OtxSdkComponent.OTX_API)
class TestResultMediaEntity:
    @staticmethod
    def default_result_media_parameters() -> dict:
        rectangle_label = LabelEntity(
            name="Rectangle Annotation Label",
            domain=Domain.DETECTION,
            color=Color(100, 200, 60),
            creation_date=datetime.datetime(year=2021, month=12, day=16),
            id=ID("rectangle_label_1"),
        )
        rectangle_annotation = Annotation(
            shape=Rectangle(x1=0.1, y1=0.4, x2=0.4, y2=0.9),
            labels=[ScoredLabel(rectangle_label)],
            id=ID("rectangle_annotation"),
        )
        annotation_scene = AnnotationSceneEntity(
            annotations=[rectangle_annotation],
            kind=AnnotationSceneKind.ANNOTATION,
            creation_date=datetime.datetime(year=2021, month=12, day=16),
            id=ID("annotation_scene"),
        )
        return {
            "name": "ResultMedia name",
            "type": "Test ResultMedia",
            "annotation_scene": annotation_scene,
            "numpy": RANDOM_IMAGE,
        }

    def optional_result_media_parameters(self) -> dict:
        optional_result_media_parameters = self.default_result_media_parameters()
        roi_label = LabelEntity(
            "ROI label",
            Domain.DETECTION,
            Color(10, 200, 40),
            creation_date=datetime.datetime(year=2021, month=12, day=18),
            id=ID("roi_label_1"),
        )
        roi = Annotation(
            shape=Rectangle(x1=0.3, y1=0.2, x2=0.7, y2=0.6),
            labels=[ScoredLabel(roi_label)],
            id=ID("roi_annotation"),
        )
        result_media_label = LabelEntity(
            "ResultMedia label",
            Domain.CLASSIFICATION,
            Color(200, 60, 100),
            creation_date=datetime.datetime(year=2021, month=12, day=20),
            id=ID("result_media_1"),
        )
        optional_result_media_parameters["roi"] = roi
        optional_result_media_parameters["label"] = result_media_label
        return optional_result_media_parameters

    def result_media(self):
        return ResultMediaEntity(**self.optional_result_media_parameters())

    @staticmethod
    def check_result_media_attributes(result_media: ResultMediaEntity, expected_values: dict):
        assert result_media.name == expected_values.get("name")
        assert result_media.type == expected_values.get("type")
        assert result_media.annotation_scene == expected_values.get("annotation_scene")
        assert np.array_equal(result_media.numpy, expected_values.get("numpy"))
        if not expected_values.get("roi"):
            assert isinstance(result_media.roi.shape, Rectangle)
            assert Rectangle.is_full_box(result_media.roi.shape)
        else:
            assert result_media.roi == expected_values.get("roi")
        assert result_media.label == expected_values.get("label")

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_result_media_initialization(self):
        """
        <b>Description:</b>
        Check ResultMediaEntity class object initialization

        <b>Input data:</b>
        ResultMediaEntity class object with specified "name", "type", "annotation_scene", "numpy", "roi"
        and "label" parameters

        <b>Expected results:</b>
        Test passes if attributes of initialized ResultMediaEntity class object are equal to expected

        <b>Steps</b>
        1. Check attributes of ResultMediaEntity class object initialized with default optional parameters
        2. Check attributes of ResultMediaEntity class object initialized with specified optional parameters
        """
        # Checking attributes of ResultMediaEntity class object initialized with default optional parameters
        initialization_params = self.default_result_media_parameters()
        result_media = ResultMediaEntity(**initialization_params)
        self.check_result_media_attributes(result_media, initialization_params)
        # Checking attributes of ResultMediaEntity class object initialized with specified optional parameters
        initialization_params = self.optional_result_media_parameters()
        result_media = ResultMediaEntity(**initialization_params)
        self.check_result_media_attributes(result_media, initialization_params)

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_result_media_repr(self):
        """
        <b>Description:</b>
        Check ResultMediaEntity class object __repr__ method

        <b>Input data:</b>
        ResultMediaEntity class object with specified "name", "type", "annotation_scene", "numpy", "roi"
        and "label" parameters

        <b>Expected results:</b>
        Test passes if value returned by __repr__ method is equal to expected

        <b>Steps</b>
        1. Check value returned by __repr__ method for ResultMediaEntity class object initialized with default optional
        parameters
        2. Check value returned by __repr__ method for ResultMediaEntity class object initialized with specified
        optional parameters
        """
        # Checking __repr__ method for ResultMediaEntity class object initialized with default optional parameters
        initialization_params = self.default_result_media_parameters()
        annotation_scene = initialization_params.get("annotation_scene")
        result_media = ResultMediaEntity(**initialization_params)
        assert repr(result_media) == (
            f"ResultMediaEntity(name=ResultMedia name, type=Test ResultMedia, annotation_scene={annotation_scene}, "
            f"roi={result_media.roi}, label=None)"
        )
        # Checking __repr__ method for ResultMediaEntity class object initialized with specified optional parameters
        initialization_params = self.optional_result_media_parameters()
        annotation_scene = initialization_params.get("annotation_scene")
        roi = initialization_params.get("roi")
        label = initialization_params.get("label")
        result_media = ResultMediaEntity(**initialization_params)
        assert repr(result_media) == (
            f"ResultMediaEntity(name=ResultMedia name, type=Test ResultMedia, annotation_scene={annotation_scene}, "
            f"roi={roi}, label={label})"
        )

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_result_media_shape(self):
        """
        <b>Description:</b>
        Check ResultMediaEntity class object "width" and "height" properties

        <b>Input data:</b>
        ResultMediaEntity class object with specified "name", "type", "annotation_scene", "numpy", "roi"
        and "label" parameters

        <b>Expected results:</b>
        Test passes if values returned by "width" and "height" properties are equal to expected

        <b>Steps</b>
        1. Check values returned by "width" and "height" properties for initialized ResultMediaEntity object
        2. Manually set new value of "numpy" property and check re-check "numpy", "width" and "height" properties
        """
        # Checking values returned by "width" and "height" properties for initialized ResultMediaEntity object
        result_media = self.result_media()
        assert result_media.width == 64
        assert result_media.height == 32
        # Manually setting new value of "numpy" property and re-checking "numpy, "width" and "height" properties
        new_numpy = np.random.uniform(low=0.0, high=255.0, size=(16, 32, 3))
        result_media.numpy = new_numpy
        np.array_equal(result_media.numpy, new_numpy)
        assert result_media.width == 32
        assert result_media.height == 16

    @pytest.mark.priority_medium
    @pytest.mark.unit
    @pytest.mark.reqids(Requirements.REQ_1)
    def test_result_media_eq(self):
        """
        <b>Description:</b>
        Check ResultMediaEntity class object __eq__ method

        <b>Input data:</b>
        ResultMediaEntity class objects with specified "name", "type", "annotation_scene", "numpy", "roi"
        and "label" parameters

        <b>Expected results:</b>
        Test passes if value returned by __eq__ method is equal to expected

        <b>Steps</b>
        1. Check value returned by __eq__ method for comparing equal ResultMediaEntity objects
        2. Check value returned by __eq__ method for comparing ResultMediaEntity objects with unequal
        "name", "type", "label" and "numpy" parameters - expected equality
        3. Check value returned by __eq__ method for comparing ResultMediaEntity objects with unequal
        "annotation_scene" and "roi" parameters - expected inequality
        4. Check value returned by __eq__ method for comparing ResultMediaEntity with different type object
        """
        initialization_params = self.optional_result_media_parameters()
        result_media = ResultMediaEntity(**initialization_params)
        # Comparing equal ResultMediaEntity objects
        equal_result_media = ResultMediaEntity(**initialization_params)
        assert result_media == equal_result_media
        # Comparing ResultMediaEntity objects with unequal "name", "type", "label" and "numpy" parameters,
        # expected equality
        unequal_values = {
            "name": "Unequal name",
            "type": "Unequal type",
            "label": LabelEntity("Unequal label", Domain.CLASSIFICATION),
            "numpy": np.random.uniform(low=0.0, high=255.0, size=(1, 2, 3)),
        }
        for key in unequal_values:
            unequal_params = dict(initialization_params)
            unequal_params[key] = unequal_values.get(key)
            equal_result_media = ResultMediaEntity(**unequal_params)
            assert result_media == equal_result_media
        # Comparing ResultMediaEntity objects with unequal "annotation_scene" and "roi" parameters, expected inequality
        unequal_values = {
            "annotation_scene": AnnotationSceneEntity(annotations=[], kind=AnnotationSceneKind.NONE),
            "roi": Rectangle.generate_full_box(),
        }
        for key in unequal_values:
            unequal_params = dict(initialization_params)
            unequal_params[key] = unequal_values.get(key)
            unequal_result_media = ResultMediaEntity(**unequal_params)
            assert result_media != unequal_result_media
        # Comparing ResultMediaEntity with different type object
        assert result_media != str



"""Report Generating for OTX CLI."""
# Copyright (C) 2023 Intel Corporation
# SPDX-License-Identifier: Apache-2.0

import sys
from collections import defaultdict
from pathlib import Path
from pprint import pformat
from typing import Any, Dict, Union

import torch

import otx
from otx.api.entities.model_template import ModelTemplate


def get_otx_report(
    model_template: ModelTemplate,
    task_config: Dict[str, Any],
    data_config: Dict[str, Dict[str, str]],
    results: Dict[str, Any],
    output_path: Union[str, Path],
):
    """Generate CLI reports."""
    dash_line = "-" * 60 + "\n\n"
    # Header
    report_str = get_otx_cli_ascii_banner()
    report_str += dash_line
    report_str += f"Current path: {Path.cwd()}\n"
    report_str += f"sys.argv: {sys.argv}\n"
    report_str += f"OTX: {otx.__version__}\n"
    # 1. Machine Environment
    report_str += sub_title_to_str("Running Environments")
    report_str += env_info_to_str()

    # 2. Task Information (Task, Train-type, Etc.)
    if model_template and task_config:
        report_str += sub_title_to_str("Template Information")
        report_str += template_info_to_str(model_template)

    # 3. Dataset Configuration
    if data_config:
        report_str += sub_title_to_str("Dataset Information")
        report_str += data_config_to_str(data_config)

    # 4. Configurations
    report_str += sub_title_to_str("Configurations")
    report_str += task_config_to_str(task_config)
    # 5. Result Summary
    report_str += sub_title_to_str("Results")
    for key, value in results.items():
        report_str += f"\t{key}: {pformat(value)}\n"

    Path(output_path).write_text(report_str, encoding="UTF-8")


def sub_title_to_str(title: str):
    """Add sub title for report."""
    dash_line = "-" * 60
    report_str = ""
    report_str += dash_line + "\n\n"
    report_str += title + "\n\n"
    report_str += dash_line + "\n"
    return report_str


def env_info_to_str():
    """Get Environments."""
    report_str = ""
    env_info = {}
    try:
        from mmcv.utils.env import collect_env

        env_info = collect_env()
        if "PyTorch compiling details" in env_info:
            env_info.pop("PyTorch compiling details")
    except ModuleNotFoundError:
        env_info["sys.platform"] = sys.platform
        env_info["Python"] = sys.version.replace("\n", "")

        cuda_available = torch.cuda.is_available()
        env_info["CUDA available"] = cuda_available

        if cuda_available:
            devices = defaultdict(list)
            for k in range(torch.cuda.device_count()):
                devices[torch.cuda.get_device_name(k)].append(str(k))
            for name, device_ids in devices.items():
                env_info["GPU " + ",".join(device_ids)] = name
        env_info["PyTorch"] = torch.__version__
    for key, value in env_info.items():
        report_str += f"\t{key}: {value}\n"
    return report_str


def template_info_to_str(model_template: ModelTemplate):
    """Get Template information."""
    report_str = ""
    for key, value in model_template.__dict__.items():
        report_str += f"\t{key}: {pformat(value)}\n"
    return report_str


def data_config_to_str(data_config: Dict[str, Dict[str, str]]):
    """Get Dataset configuration."""
    report_str = ""
    for subset_key, subset_value in data_config.items():
        report_str += f"{subset_key}:\n"
        for key, value in subset_value.items():
            report_str += f"\t{key}: {value}\n"
    return report_str


def task_config_to_str(task_config: Dict[str, Any]):
    """Get Task configuration."""
    report_str = ""
    not_target = ["log_config"]
    for target, value in task_config.items():
        # Remove otx_dataset from the report as it is unnecessary.
        if target == "data" and isinstance(value, dict):
            for item in value.values():
                if isinstance(item, dict) and "otx_dataset" in item:
                    del item["otx_dataset"]

        if target not in not_target:
            report_str += target + ": "
            model_str = pformat(value)
            report_str += model_str + "\n"
    return report_str


def get_otx_cli_ascii_banner():
    """Get OTX ASCII banner."""
    return """

 ██████╗     ████████╗    ██╗  ██╗
██╔═══██╗    ╚══██╔══╝    ╚██╗██╔╝
██║   ██║       ██║        ╚███╔╝
██║   ██║       ██║        ██╔██╗
╚██████╔╝       ██║       ██╔╝ ██╗
 ╚═════╝        ╚═╝       ╚═╝  ╚═╝

"""
